import unittest

import torch

from gpytorch import settings
from gpytorch import ExactMarginalLogLikelihood
from gpytorch.distributions import MultivariateNormal, MultitaskMultivariateNormal
from gpytorch.kernels import ScaleKernel, RBFKernel, MultitaskKernel
from gpytorch.likelihoods import GaussianLikelihood, Likelihood, MultitaskGaussianLikelihood
from gpytorch.means import ConstantMean, MultitaskMean
from gpytorch.mlls import PredictiveLogLikelihood, MarginalLogLikelihood, VariationalELBO
from gpytorch.models import ExactGP, VariationalGP, GP
from gpytorch.test.base_test_case import BaseTestCase
from gpytorch.utils.memoize import clear_cache_hook
from gpytorch.variational import CholeskyVariationalDistribution, LMCVariationalStrategy, VariationalStrategy


class SingleGPModel(ExactGP):
    def __init__(self, train_inputs, train_targets, likelihood, batch_shape):
        super(SingleGPModel, self).__init__(train_inputs, train_targets, likelihood)
        self.mean_module = ConstantMean(batch_shape=batch_shape)
        self.covar_module = ScaleKernel(RBFKernel(batch_shape=batch_shape))

    def forward(self, x):
        mean_x = self.mean_module(x)
        covar_x = self.covar_module(x)
        return MultivariateNormal(mean_x, covar_x)


class MultitaskGPModel(ExactGP):
    def __init__(self, train_inputs, train_targets, likelihood, num_tasks):
        super(MultitaskGPModel, self).__init__(train_inputs, train_targets, likelihood)
        self.mean_module = MultitaskMean(ConstantMean(), num_tasks=num_tasks)
        self.covar_module = MultitaskKernel(ScaleKernel(RBFKernel()), num_tasks=num_tasks)

    def forward(self, x):
        mean_x = self.mean_module(x)
        covar_x = self.covar_module(x)
        return MultitaskMultivariateNormal(mean_x, covar_x)


class MultitaskVariationalGPModel(VariationalGP):
    def __init__(self, num_latents, num_tasks):
        inducing_points = torch.rand(num_latents, 21, 1)
        variational_distribution = CholeskyVariationalDistribution(
            inducing_points.size(-2), batch_shape=torch.Size([num_latents])
        )
        variational_strategy = LMCVariationalStrategy(
            VariationalStrategy(
                self, inducing_points, variational_distribution, learn_inducing_locations=True
            ),
            num_tasks=num_tasks,
            num_latents=num_latents,
            latent_dim=-1
        )
        super().__init__(variational_strategy)
        self.mean_module = ConstantMean(batch_shape=torch.Size([num_latents]))
        self.covar_module = ScaleKernel(
            RBFKernel(batch_shape=torch.Size([num_latents])),
            batch_shape=torch.Size([num_latents])
        )

    def forward(self, x):
        mean_x = self.mean_module(x)
        covar_x = self.covar_module(x)
        return MultivariateNormal(mean_x, covar_x)


class TestMissingData(BaseTestCase, unittest.TestCase):
    seed = 20
    warning = "Observation NaN policy 'fill' makes the kernel matrix dense during exact prediction."

    def _check(
        self,
        model: GP,
        likelihood: Likelihood,
        train_x: torch.Tensor,
        train_y: torch.Tensor,
        test_x: torch.Tensor,
        test_y: torch.Tensor,
        optimizer: torch.optim.Optimizer,
        mll: MarginalLogLikelihood,
        epochs: int = 30,
        atol: float = 0.2
    ) -> None:
        model.train()
        likelihood.train()

        for _ in range(epochs):
            optimizer.zero_grad()
            output = model(train_x)
            loss = -mll(output, train_y).sum()
            self.assertFalse(torch.any(torch.isnan(output.mean)).item())
            self.assertFalse(torch.any(torch.isnan(output.covariance_matrix)).item())
            self.assertFalse(torch.isnan(loss).item())
            loss.backward()
            optimizer.step()

        model.eval()
        likelihood.eval()

        with torch.no_grad():
            if isinstance(model, ExactGP):
                self._check_predictions_exact_gp(model, test_x, test_y, atol)
            else:
                prediction = model(test_x)
                self._check_prediction(prediction, test_y, atol)

    def _check_predictions_exact_gp(self, model: ExactGP, test_x: torch.Tensor, test_y: torch.Tensor, atol: float):
        with settings.observation_nan_policy("mask"):
            prediction = model(test_x)
            self._check_prediction(prediction, test_y, atol)

        clear_cache_hook(model.prediction_strategy)

        with settings.observation_nan_policy("fill"), self.assertWarns(RuntimeWarning, msg=self.warning):
            prediction = model(test_x)
            self._check_prediction(prediction, test_y, atol)

        clear_cache_hook(model.prediction_strategy)

        with settings.observation_nan_policy("mask"):
            model(test_x)
        with settings.observation_nan_policy("fill"), self.assertWarns(RuntimeWarning, msg=self.warning):
            prediction = model(test_x)
            self._check_prediction(prediction, test_y, atol)

        clear_cache_hook(model.prediction_strategy)

        with settings.observation_nan_policy("fill"), self.assertWarns(RuntimeWarning, msg=self.warning):
            model(test_x)
        with settings.observation_nan_policy("mask"):
            prediction = model(test_x)
            self._check_prediction(prediction, test_y, atol)

    def _check_prediction(self, prediction: MultivariateNormal, test_y: torch.Tensor, atol: float):
        self.assertFalse(torch.any(torch.isnan(prediction.mean)).item())
        self.assertFalse(torch.any(torch.isnan(prediction.covariance_matrix)).item())
        self.assertAllClose(prediction.mean, test_y, atol=atol)

    def test_single(self):
        train_x = torch.linspace(0, 1, 41)
        test_x = torch.linspace(0, 1, 51)
        train_y = torch.sin(2 * torch.pi * train_x).squeeze()
        train_y += torch.normal(0, 0.01, train_y.shape)
        test_y = torch.sin(2 * torch.pi * test_x).squeeze()
        train_y[::4] = torch.nan

        batch_shape = torch.Size(())
        likelihood = GaussianLikelihood(batch_shape=batch_shape)
        model = SingleGPModel(train_x, train_y, likelihood, batch_shape=batch_shape)

        mll = ExactMarginalLogLikelihood(likelihood, model)
        optimizer = torch.optim.Adam(model.parameters(), lr=0.15)

        with settings.observation_nan_policy("mask"):
            self._check(model, likelihood, train_x, train_y, test_x, test_y, optimizer, mll)

    def test_single_batch(self):
        train_x = torch.stack([torch.linspace(0, 1, 41), torch.linspace(1, 2, 41)]).reshape(2, 41, 1)
        test_x = torch.stack([torch.linspace(0, 1, 51), torch.linspace(1, 2, 51)]).reshape(2, 51, 1)
        train_y = torch.sin(2 * torch.pi * train_x).squeeze()
        train_y += torch.normal(0, 0.01, train_y.shape)
        test_y = torch.sin(2 * torch.pi * test_x).squeeze()
        train_y[0, ::4] = torch.nan

        batch_shape = torch.Size((2,))
        likelihood = GaussianLikelihood(batch_shape=batch_shape)
        model = SingleGPModel(train_x, train_y, likelihood, batch_shape=batch_shape)

        mll = ExactMarginalLogLikelihood(likelihood, model)
        optimizer = torch.optim.Adam(model.parameters(), lr=0.15)

        with settings.observation_nan_policy("mask"):
            self._check(model, likelihood, train_x, train_y, test_x, test_y, optimizer, mll)

    def test_multitask(self):
        num_tasks = 10
        train_x = torch.linspace(0, 1, 41)
        test_x = torch.linspace(0, 1, 51)
        coefficients = torch.rand(1, num_tasks)
        train_y = torch.sin(2 * torch.pi * train_x)[:, None] * coefficients
        train_y += torch.normal(0, 0.01, train_y.shape)
        test_y = torch.sin(2 * torch.pi * test_x)[:, None] * coefficients
        train_y[::3, : num_tasks // 2] = torch.nan
        train_y[::4, num_tasks // 2 :] = torch.nan

        likelihood = MultitaskGaussianLikelihood(num_tasks)
        model = MultitaskGPModel(train_x, train_y, likelihood, num_tasks)

        mll = ExactMarginalLogLikelihood(likelihood, model)
        optimizer = torch.optim.Adam(model.parameters(), lr=0.15)

        with settings.observation_nan_policy("mask"):
            self._check(model, likelihood, train_x, train_y, test_x, test_y, optimizer, mll)

    def test_variational_multitask(self):
        num_latents = 1
        train_x = torch.linspace(0, 1, 41)
        test_x = torch.linspace(0, 1, 51)
        train_y = torch.stack([
            torch.sin(train_x * (2 * torch.pi)) + torch.randn(train_x.size()) * 0.2,
            -torch.sin(train_x * (2 * torch.pi)) + torch.randn(train_x.size()) * 0.2,
        ], -1)
        test_y = torch.stack([
            torch.sin(test_x * (2 * torch.pi)),
            -torch.sin(test_x * (2 * torch.pi)),
        ], -1)
        num_tasks = train_y.shape[-1]

        # nan out a few train_y
        train_y[-3:, 1] = torch.nan

        likelihood = MultitaskGaussianLikelihood(num_tasks=num_tasks)
        model = MultitaskVariationalGPModel(num_latents, num_tasks)
        model.train()
        likelihood.train()

        optimizer = torch.optim.Adam([
            {'params': model.parameters()},
            {'params': likelihood.parameters()},
        ], lr=0.05)

        mll = VariationalELBO(likelihood, model, num_data=train_y.size(0))
        with settings.observation_nan_policy("mask"):
            self._check(model, likelihood, train_x, train_y, test_x, test_y, optimizer, mll, epochs=50, atol=0.7)
        with settings.observation_nan_policy("fill"):
            self._check(model, likelihood, train_x, train_y, test_x, test_y, optimizer, mll, epochs=50, atol=0.3)

        mll = PredictiveLogLikelihood(likelihood, model, num_data=train_y.size(0))
        with settings.observation_nan_policy("mask"):
            self._check(model, likelihood, train_x, train_y, test_x, test_y, optimizer, mll, epochs=50, atol=0.7)
        with settings.observation_nan_policy("fill"):
            self._check(model, likelihood, train_x, train_y, test_x, test_y, optimizer, mll, epochs=50, atol=0.3)



#!/usr/bin/env python3

import unittest

from torch import Tensor

from gpytorch.priors import GammaPrior, HalfCauchyPrior, LogNormalPrior, NormalPrior


TRANSFORMED_ERROR_MSG = """Priors of TransformedDistributions should not have their \
'_transformed' attributes modified, these are just copies of the base attribute. \
Please modify the base attribute (e.g. {}) instead."""


class TestPrior(unittest.TestCase):
    def test_state_dict(self):
        normal = NormalPrior(0.1, 1).state_dict()
        self.assertTrue("loc" in normal)
        self.assertTrue("scale" in normal)
        self.assertEqual(normal["loc"], 0.1)

        gamma = GammaPrior(1.1, 2).state_dict()
        self.assertTrue("concentration" in gamma)
        self.assertTrue("rate" in gamma)
        self.assertEqual(gamma["concentration"], 1.1)

        ln = LogNormalPrior(2.1, 1.2).state_dict()
        self.assertTrue("_transformed_loc" in ln)
        self.assertTrue("_transformed_scale" in ln)
        self.assertEqual(ln["_transformed_loc"], 2.1)

        hc = HalfCauchyPrior(1.3).state_dict()
        self.assertTrue("_transformed_scale" in hc)

    def test_load_state_dict(self):
        ln1 = LogNormalPrior(loc=0.5, scale=0.1)
        ln2 = LogNormalPrior(loc=2.5, scale=2.1)
        gm1 = GammaPrior(concentration=0.5, rate=0.1)
        gm2 = GammaPrior(concentration=2.5, rate=2.1)
        hc1 = HalfCauchyPrior(scale=1.1)
        hc2 = HalfCauchyPrior(scale=101.1)

        ln2.load_state_dict(ln1.state_dict())
        self.assertEqual(ln2.loc, ln1.loc)
        self.assertEqual(ln2.scale, ln1.scale)

        gm2.load_state_dict(gm1.state_dict())
        self.assertEqual(gm2.concentration, gm1.concentration)
        self.assertEqual(gm2.rate, gm1.rate)

        hc2.load_state_dict(hc1.state_dict())
        self.assertEqual(hc2.scale, hc1.scale)

    def test_transformed_attributes(self):
        norm = NormalPrior(loc=2.5, scale=2.1)
        ln = LogNormalPrior(loc=2.5, scale=2.1)
        hc = HalfCauchyPrior(scale=2.2)

        with self.assertRaisesRegex(AttributeError, "'NormalPrior' object has no attribute '_transformed_loc'"):
            getattr(norm, "_transformed_loc")

        self.assertTrue(getattr(ln, "_transformed_loc"), 2.5)
        norm.loc = Tensor([1.01])
        ln.loc = Tensor([1.01])
        self.assertEqual(ln._transformed_loc, 1.01)
        with self.assertRaises(AttributeError):
            ln._transformed_loc = 1.1

        with self.assertRaises(AttributeError):
            hc._transformed_scale = 1.01



#!/usr/bin/env python3

from typing import Optional, Tuple

import torch
from torch import Tensor

from ..constraints import Interval, Positive
from ..priors import Prior
from .kernel import Kernel


class ConstantKernel(Kernel):
    """
    Constant covariance kernel for the probabilistic inference of constant coefficients.

    ConstantKernel represents the prior variance `k(x1, x2) = var(c)` of a constant `c`.
    The prior variance of the constant is optimized during the GP hyper-parameter
    optimization stage. The actual value of the constant is computed (implicitly) using
    the linear algebraic approaches for the computation of GP samples and posteriors.

    The constant kernel `k_constant` is most useful as a modification of an arbitrary
    base kernel `k_base`:
    1) Additive constants: The modification `k_base + k_constant` allows the GP to
    infer a non-zero asymptotic value far from the training data, which generally
    leads to more accurate extrapolation. Notably, the uncertainty in this constant
    value affects the posterior covariances through the posterior inference equations.
    This is not the case when a constant prior mean is not used, since the prior mean
    does not show up the posterior covariance and is regularized by the log-determinant
    during the optimization of the marginal likelihood.
    2) Multiplicative constants: The modification `k_base * k_constant` allows the GP to
    modulate the variance of the kernel `k_base`, and is mathematically identical to
    `ScaleKernel(base_kernel)` with the same constant.
    """

    has_lengthscale = False

    def __init__(
        self,
        batch_shape: Optional[torch.Size] = None,
        constant_prior: Optional[Prior] = None,
        constant_constraint: Optional[Interval] = None,
        active_dims: Optional[Tuple[int, ...]] = None,
    ):
        """Constructor of ConstantKernel.

        Args:
            batch_shape: The batch shape of the kernel.
            constant_prior: Prior over the constant parameter.
            constant_constraint: Constraint to place on constant parameter.
            active_dims: The dimensions of the input with which to evaluate the kernel.
                This is mute for the constant kernel, but added for compatability with
                the Kernel API.
        """
        super().__init__(batch_shape=batch_shape, active_dims=active_dims)

        self.register_parameter(
            name="raw_constant",
            parameter=torch.nn.Parameter(torch.zeros(*self.batch_shape, 1)),
        )

        if constant_prior is not None:
            if not isinstance(constant_prior, Prior):
                raise TypeError("Expected gpytorch.priors.Prior but got " + type(constant_prior).__name__)
            self.register_prior(
                "constant_prior",
                constant_prior,
                lambda m: m.constant,
                lambda m, v: m._set_constant(v),
            )

        if constant_constraint is None:
            constant_constraint = Positive()
        self.register_constraint("raw_constant", constant_constraint)

    @property
    def constant(self) -> Tensor:
        return self.raw_constant_constraint.transform(self.raw_constant)

    @constant.setter
    def constant(self, value: Tensor) -> None:
        self._set_constant(value)

    def _set_constant(self, value: Tensor) -> None:
        value = value.view(*self.batch_shape, 1)
        self.initialize(raw_constant=self.raw_constant_constraint.inverse_transform(value))

    def forward(
        self,
        x1: Tensor,
        x2: Tensor,
        diag: Optional[bool] = False,
        last_dim_is_batch: Optional[bool] = False,
    ) -> Tensor:
        """Evaluates the constant kernel.

        Args:
            x1: First input tensor of shape (batch_shape x n1 x d).
            x2: Second input tensor of shape (batch_shape x n2 x d).
            diag: If True, returns the diagonal of the covariance matrix.
            last_dim_is_batch: If True, the last dimension of size `d` of the input
                tensors are treated as a batch dimension.

        Returns:
            A (batch_shape x n1 x n2)-dim, resp. (batch_shape x n1)-dim, tensor of
            constant covariance values if diag is False, resp. True.
        """
        if last_dim_is_batch:
            x1 = x1.transpose(-1, -2).unsqueeze(-1)
            x2 = x2.transpose(-1, -2).unsqueeze(-1)

        dtype = torch.promote_types(x1.dtype, x2.dtype)
        batch_shape = torch.broadcast_shapes(x1.shape[:-2], x2.shape[:-2])
        shape = batch_shape + (x1.shape[-2],) + (() if diag else (x2.shape[-2],))
        constant = self.constant.to(dtype=dtype, device=x1.device)

        if not diag:
            constant = constant.unsqueeze(-1)

        if last_dim_is_batch:
            constant = constant.unsqueeze(-1)

        return constant.expand(shape)



#!/usr/bin/env python3

import math
import os
import random
import unittest
import warnings
from math import pi
from unittest.mock import patch

import linear_operator
import torch
from torch import optim

import gpytorch
from gpytorch.constraints import GreaterThan
from gpytorch.distributions import MultivariateNormal
from gpytorch.kernels import RFFKernel, ScaleKernel
from gpytorch.likelihoods import GaussianLikelihood
from gpytorch.means import ConstantMean
from gpytorch.priors import SmoothedBoxPrior
from gpytorch.utils.warnings import NumericalWarning


class RFFRegressionModel(gpytorch.models.ExactGP):
    def __init__(self, train_x, train_y, num_features: int):
        assert num_features % 2 == 0, "num_features must be even"
        likelihood = GaussianLikelihood()
        super().__init__(train_x, train_y, likelihood)
        self.mean_module = ConstantMean(constant_prior=SmoothedBoxPrior(-1e-5, 1e-5))
        self.covar_module = ScaleKernel(RFFKernel(num_samples=(num_features // 2)))

    def forward(self, x):
        mean_x = self.mean_module(x)
        covar_x = self.covar_module(x)
        return MultivariateNormal(mean_x, covar_x)


class LinearRegressionModel(gpytorch.models.ExactGP):
    def __init__(self, train_x, train_y):
        # We throw away num_features here, because it corresponds to the number of features in the data
        likelihood = GaussianLikelihood(noise_constraint=GreaterThan(1e-3))
        super().__init__(train_x, train_y, likelihood)
        self.mean_module = ConstantMean(constant_prior=SmoothedBoxPrior(-1e-5, 1e-5))
        self.covar_module = ScaleKernel(gpytorch.kernels.LinearKernel())

    def forward(self, x):
        mean_x = self.mean_module(x)
        covar_x = self.covar_module(x)
        return MultivariateNormal(mean_x, covar_x)


class _AbstractTestLowRankRegression():
    num_features = 20  # Expected rank of the covariance matrix

    def setUp(self):
        if os.getenv("UNLOCK_SEED") is None or os.getenv("UNLOCK_SEED").lower() == "false":
            self.rng_state = torch.get_rng_state()
            torch.manual_seed(0)
            random.seed(0)

    def tearDown(self):
        if hasattr(self, "rng_state"):
            torch.set_rng_state(self.rng_state)

    def make_data(self):
        raise NotImplementedError

    def make_model(self, train_x, train_y):
        raise NotImplementedError

    def test_rff_mean_abs_error(self):
        # Suppress numerical warnings
        warnings.simplefilter("ignore", NumericalWarning)

        # Track all matrix sizes passed to cholesky_ex
        cholesky_sizes = []
        original_cholesky_ex = torch.linalg.cholesky_ex

        def wrapped_cholesky_ex(matrix, *args, **kwargs):
            cholesky_sizes.append(matrix.shape)
            return original_cholesky_ex(matrix, *args, **kwargs)

        with (
            patch(
                "torch.linalg.cholesky_ex",
                wraps=wrapped_cholesky_ex,
            ) as mocked_cholesky_ex,
        ):
            train_x, train_y, test_x, test_y = self.make_data()
            gp_model = self.make_model(train_x, train_y)
            mll = gpytorch.mlls.ExactMarginalLogLikelihood(gp_model.likelihood, gp_model)

            # Optimize the model
            gp_model.train()

            optimizer = optim.Adam(gp_model.parameters(), lr=0.1)
            for i in range(30):
                calls_before = mocked_cholesky_ex.call_count

                optimizer.zero_grad()
                output = gp_model(train_x)
                loss = -mll(output, train_y)
                loss.backward()
                optimizer.step()

                calls_after = mocked_cholesky_ex.call_count

                # Assert exactly one call per iteration
                self.assertEqual(
                    calls_after - calls_before,
                    1,
                    f"Expected 1 cholesky call in iteration {i}, got {calls_after - calls_before}",
                )

                # Check that we have the right LinearOperator type
                kernel = gp_model.likelihood(gp_model(train_x)).lazy_covariance_matrix.evaluate_kernel()
                if train_y.size(-1) >= self.num_features:
                    self.assertIsInstance(kernel, linear_operator.operators.LowRankRootAddedDiagLinearOperator)
                else:
                    self.assertIsInstance(kernel, linear_operator.operators.AddedDiagLinearOperator)

            for param in gp_model.parameters():
                self.assertTrue(param.grad is not None)
                self.assertGreater(param.grad.norm().item(), 0)

            # Test the model
            gp_model.eval()

            with torch.no_grad():
                calls_before = mocked_cholesky_ex.call_count

                test_dist = gp_model(test_x)
                _ = test_dist.covariance_matrix  # Force computation of covariances
                _ = test_dist.rsample(torch.Size((1000,)))  # Force sampling
                mean_abs_error = torch.mean(torch.abs(test_y - test_dist.mean))

                calls_after = mocked_cholesky_ex.call_count

                # Assert exactly two calls (mean/covar + sampling) during prediction
                self.assertEqual(
                    calls_after - calls_before,
                    2,
                    f"Expected 2 cholesky calls during prediction, got {calls_after - calls_before}",
                )

        # Check that the model makes good predictions
        self.assertLess(mean_abs_error.squeeze().item(), 0.15)

        # Check sizes of cholesky systems during training
        for idx, size in enumerate(cholesky_sizes[:-2]):
            expected_size = min(self.num_features, train_y.size(-1))
            self.assertEqual(
                size,
                torch.Size([expected_size, expected_size]),
                f"Expected tensor of size [{expected_size} x {expected_size}] for prediction call {idx}, got {size}",
            )
        # Check sizes of first prediction Cholesky system
        expected_size = min(self.num_features, train_y.size(-1))
        self.assertEqual(
            cholesky_sizes[-2],
            torch.Size([expected_size, expected_size]),
            f"Expected tensor of size [{expected_size} x {expected_size}] for prediction call {idx}, got {size}",
        )
        # Check sizes of first prediction Cholesky system
        expected_size = self.num_features
        self.assertEqual(
            cholesky_sizes[-1],
            torch.Size([expected_size, expected_size]),
            f"Expected tensor of size [{expected_size} x {expected_size}] for prediction call {idx}, got {size}",
        )


class TestRFFRegression(_AbstractTestLowRankRegression, unittest.TestCase):
    def make_data(self):
        train_x = torch.linspace(0, 1, 100)
        train_y = torch.sin(train_x * (2 * pi))
        train_y.add_(torch.randn_like(train_y), alpha=1e-2)
        test_x = torch.rand(51)
        test_y = torch.sin(test_x * (2 * pi))
        return train_x, train_y, test_x, test_y

    def make_model(self, train_x, train_y):
        return RFFRegressionModel(train_x, train_y, num_features=self.num_features)


class TestLinearRegressionSmallD(_AbstractTestLowRankRegression, unittest.TestCase):
    def make_data(self):
        # We throw away num_features here, because it corresponds to the number of RFF features,
        # not the number of dimensions in the data.
        train_x = torch.randn(100, self.num_features)
        beta = torch.randn(self.num_features)
        train_y = train_x @ beta + torch.randn_like(train_x[..., 0]) * 0.1
        test_x = torch.randn(51, self.num_features)
        test_y = test_x @ beta
        return train_x, train_y, test_x, test_y

    def make_model(self, train_x, train_y):
        return LinearRegressionModel(train_x, train_y)


class TestLinearRegressionLargeD(_AbstractTestLowRankRegression, unittest.TestCase):
    num_features = 200  # Expected # of features

    def make_data(self):
        # We throw away num_features here, because it corresponds to the number of RFF features,
        # not the number of dimensions in the data.
        feature_sizes = torch.cat([torch.ones(10), torch.full((self.num_features - 10,), 1e-4)])
        train_x = torch.randn(100, self.num_features) * feature_sizes
        beta = torch.randn(self.num_features) / math.sqrt(10)
        train_y = train_x @ beta + torch.randn_like(train_x[..., 0]) * 0.01
        test_x = torch.randn(51, self.num_features) * feature_sizes
        test_y = test_x  @ beta
        return train_x, train_y, test_x, test_y

    def make_model(self, train_x, train_y):
        return LinearRegressionModel(train_x, train_y)



#!/usr/bin/env python3

import unittest

from gpytorch.kernels import PeriodicKernel as GPeriodicKernel
from gpytorch.kernels.keops import PeriodicKernel
from gpytorch.priors import NormalPrior
from gpytorch.test.base_keops_test_case import BaseKeOpsTestCase
from gpytorch.test.base_kernel_test_case import BaseKernelTestCase

try:
    import pykeops  # noqa

    class TestPeriodicKeOpsBaseKernel(unittest.TestCase, BaseKernelTestCase):
        def create_kernel_no_ard(self, **kwargs):
            return PeriodicKernel(**kwargs)

        def create_kernel_ard(self, num_dims, **kwargs):
            return PeriodicKernel(ard_num_dims=num_dims, **kwargs)

    class TestPeriodicKeOpsKernel(unittest.TestCase, BaseKeOpsTestCase):
        @property
        def k1(self):
            return PeriodicKernel

        @property
        def k2(self):
            return GPeriodicKernel

    def create_kernel_with_prior(self, period_length_prior):
        return self.k1(period_length_prior=period_length_prior)

    def test_prior_type(self):
        """
        Raising TypeError if prior type is other than gpytorch.priors.Prior
        """
        kernel_fn = lambda prior: self.create_kernel_with_prior(prior)
        kernel_fn(None)
        kernel_fn(NormalPrior(0, 1))
        self.assertRaises(TypeError, kernel_fn, 1)

except ImportError:
    pass

if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3

import math
import unittest
import warnings
from pathlib import Path
from unittest.mock import MagicMock, patch

import linear_operator
import torch
from torch import optim

import gpytorch
from gpytorch.likelihoods import GaussianLikelihood
from gpytorch.models import ApproximateGP
from gpytorch.test.base_test_case import BaseTestCase
from gpytorch.test.utils import least_used_cuda_device
from gpytorch.utils.warnings import OldVersionWarning


def train_data(cuda=False):
    train_x = torch.linspace(0, 1, 260)
    train_y = torch.cos(train_x * (2 * math.pi))
    if cuda:
        return train_x.cuda(), train_y.cuda()
    else:
        return train_x, train_y


class SVGPRegressionModel(ApproximateGP):
    def __init__(self, inducing_points, distribution_cls):
        variational_distribution = distribution_cls(inducing_points.size(-1))
        variational_strategy = gpytorch.variational.VariationalStrategy(
            self, inducing_points, variational_distribution, learn_inducing_locations=True, jitter_val=1e-4
        )
        super(SVGPRegressionModel, self).__init__(variational_strategy)
        self.mean_module = gpytorch.means.ConstantMean()
        self.covar_module = gpytorch.kernels.ScaleKernel(gpytorch.kernels.RBFKernel())

    def forward(self, x):
        mean_x = self.mean_module(x)
        covar_x = self.covar_module(x)
        latent_pred = gpytorch.distributions.MultivariateNormal(mean_x, covar_x)
        return latent_pred


class TestSVGPRegression(BaseTestCase, unittest.TestCase):
    seed = 0

    def test_loading_old_model(self):
        train_x, train_y = train_data(cuda=False)
        likelihood = GaussianLikelihood()
        model = SVGPRegressionModel(torch.linspace(0, 1, 25), gpytorch.variational.CholeskyVariationalDistribution)
        data_file = Path(__file__).parent.joinpath("old_variational_strategy_model.pth").resolve()
        state_dicts = torch.load(data_file)
        likelihood.load_state_dict(state_dicts["likelihood"], strict=False)

        # Ensure we get a warning
        with warnings.catch_warnings(record=True) as ws:
            warnings.simplefilter("always", OldVersionWarning)

            model.load_state_dict(state_dicts["model"])
            self.assertTrue(any(issubclass(w.category, OldVersionWarning) for w in ws))

        with torch.no_grad():
            model.eval()
            likelihood.eval()
            test_preds = likelihood(model(train_x)).mean.squeeze()
            mean_abs_error = torch.mean(torch.abs(train_y - test_preds) / 2)
            self.assertLess(mean_abs_error.item(), 1e-1)

    def test_regression_error(
        self,
        cuda=False,
        mll_cls=gpytorch.mlls.VariationalELBO,
        distribution_cls=gpytorch.variational.CholeskyVariationalDistribution,
    ):
        train_x, train_y = train_data(cuda=cuda)
        likelihood = GaussianLikelihood()
        model = SVGPRegressionModel(torch.linspace(0, 1, 25), distribution_cls)
        mll = mll_cls(likelihood, model, num_data=len(train_y))
        if cuda:
            likelihood = likelihood.cuda()
            model = model.cuda()
            mll = mll.cuda()

        # Find optimal model hyperparameters
        model.train()
        likelihood.train()
        optimizer = optim.Adam([{"params": model.parameters()}, {"params": likelihood.parameters()}], lr=0.01)

        _wrapped_cg = MagicMock(wraps=linear_operator.utils.linear_cg)
        _cg_mock = patch("linear_operator.utils.linear_cg", new=_wrapped_cg)
        with _cg_mock as cg_mock:
            for _ in range(150):
                optimizer.zero_grad()
                output = model(train_x)
                loss = -mll(output, train_y)
                loss.backward()
                optimizer.step()

            for param in model.parameters():
                self.assertTrue(param.grad is not None)
                self.assertGreater(param.grad.norm().item(), 0)
            for param in likelihood.parameters():
                self.assertTrue(param.grad is not None)
                self.assertGreater(param.grad.norm().item(), 0)

            # Set back to eval mode
            model.eval()
            likelihood.eval()
            test_preds = likelihood(model(train_x)).mean.squeeze()
            mean_abs_error = torch.mean(torch.abs(train_y - test_preds) / 2)
            self.assertLess(mean_abs_error.item(), 1e-1)

            # Make sure CG was called (or not), and no warnings were thrown
            self.assertFalse(cg_mock.called)

            if distribution_cls is gpytorch.variational.CholeskyVariationalDistribution:
                # finally test fantasization
                # we only will check that tossing the entire training set into the model will reduce the mae
                model.likelihood = likelihood
                fant_model = model.get_fantasy_model(train_x, train_y)
                fant_preds = fant_model.likelihood(fant_model(train_x)).mean.squeeze()
                updated_abs_error = torch.mean(torch.abs(train_y - fant_preds) / 2)
                # TODO: figure out why this error is worse than before
                self.assertLess(updated_abs_error.item(), 0.15)

    def test_predictive_ll_regression_error(self):
        return self.test_regression_error(
            mll_cls=gpytorch.mlls.PredictiveLogLikelihood,
            distribution_cls=gpytorch.variational.MeanFieldVariationalDistribution,
        )

    def test_predictive_ll_regression_error_delta(self):
        return self.test_regression_error(
            mll_cls=gpytorch.mlls.PredictiveLogLikelihood,
            distribution_cls=gpytorch.variational.DeltaVariationalDistribution,
        )

    def test_robust_regression_error(self):
        return self.test_regression_error(mll_cls=gpytorch.mlls.GammaRobustVariationalELBO)

    def test_regression_error_cuda(self):
        if not torch.cuda.is_available():
            return
        with least_used_cuda_device():
            return self.test_regression_error(cuda=True)


if __name__ == "__main__":
    unittest.main()



import math
import unittest

import torch

import gpytorch
from gpytorch.distributions import MultitaskMultivariateNormal, MultivariateNormal
from gpytorch.kernels import MultitaskKernel, RBFKernel, ScaleKernel
from gpytorch.likelihoods import GaussianLikelihood, MultitaskGaussianLikelihood
from gpytorch.means import ConstantMean, MultitaskMean
from gpytorch.metrics import (  # average_coverage_error,
    mean_absolute_error,
    mean_squared_error,
    mean_standardized_log_loss,
    negative_log_predictive_density,
    quantile_coverage_error,
    standardized_mean_squared_error,
)
from gpytorch.models import ExactGP


class ExactGPModel(ExactGP):
    def __init__(self, train_x, train_y, likelihood):
        super().__init__(train_x, train_y, likelihood)
        self.mean_module = ConstantMean()
        self.covar_module = ScaleKernel(RBFKernel())

    def forward(self, x):
        mean_x = self.mean_module(x)
        covar_x = self.covar_module(x)
        return MultivariateNormal(mean_x, covar_x)


class ExactMultiTaskGPModel(ExactGP):
    def __init__(self, train_x, train_y, likelihood):
        super().__init__(train_x, train_y, likelihood)
        self.mean_module = MultitaskMean(ConstantMean(), num_tasks=2)
        self.covar_module = MultitaskKernel(RBFKernel(), num_tasks=2, rank=1)

    def forward(self, x):
        mean_x = self.mean_module(x)
        covar_x = self.covar_module(x)
        return MultitaskMultivariateNormal(mean_x, covar_x)


class TestMetricsSingleTask(unittest.TestCase):
    def setUp(self, batch_shape=None):
        train_x, train_y = self.create_data_and_labels(batch_shape=batch_shape)
        untrained_model = self.create_model(train_x, train_y)
        self.trained_model = self.train_model(untrained_model, train_x, train_y)
        self.untrained_model = self.create_model(train_x, train_y)
        self.testing_pts = self.create_data_and_labels(batch_shape=batch_shape, is_training=False)
        self.train_x = train_x
        self.train_y = train_y

    def create_model(self, train_x, train_y):
        likelihood = GaussianLikelihood(batch_shape=train_x.shape[:-2])
        model = ExactGPModel(train_x, train_y, likelihood)
        # assume determinisim
        return model

    def create_data_and_labels(self, n=20, batch_shape=None, seed=0, is_training=True):
        input_range = [0.0, 1.0] if is_training else [1.0, 1.2]
        torch.random.manual_seed(seed)
        if batch_shape is None:
            batch_shape = torch.Size()
        inputs = torch.linspace(*input_range, n).view(-1, 1).contiguous()
        if batch_shape != torch.Size():
            inputs = inputs.unsqueeze(0)
            inputs = inputs.repeat(*batch_shape, 1, 1)
        labels = torch.sin(inputs * (2 * math.pi)) + torch.randn_like(inputs) * 0.1
        return inputs, labels.squeeze()

    def get_metric(self, model, likelihood, train_data, labels, metric, **kwargs):
        model.eval()
        likelihood.eval()
        pred_dist = likelihood(model(train_data))
        return metric(pred_dist, labels, **kwargs)

    def train_model(self, model, train_x, train_y):
        mll = gpytorch.mlls.ExactMarginalLogLikelihood(model.likelihood, model)
        optimizer = torch.optim.Adam(model.parameters(), lr=0.1)
        model.train()
        model.likelihood.train()
        for _ in range(10):
            optimizer.zero_grad()
            output = model(train_x)
            loss = -mll(output, train_y).sum()
            loss.backward()
            optimizer.step()

        return model

    def check_metric(self, metric, **kwargs):
        init_value = self.get_metric(
            self.untrained_model,
            self.untrained_model.likelihood,
            *self.testing_pts,
            metric,
            **kwargs,
        )
        final_value = self.get_metric(
            self.trained_model,
            self.trained_model.likelihood,
            *self.testing_pts,
            metric,
            **kwargs,
        )
        return init_value, final_value

    def _test_metric(self, metric, should_check_differentiable=True, **kwargs):
        init_value, final_value = self.check_metric(metric, **kwargs)
        self.assertEqual(final_value.shape, self.train_y.shape[:-1])
        self.assertTrue(torch.all(final_value <= init_value))
        if should_check_differentiable:
            self.assertTrue(final_value.requires_grad)  # check that the metric is differentiable

    def test_mean_absolute_error(self):
        self._test_metric(mean_absolute_error)

    def test_mean_squared_error(self):
        self._test_metric(mean_squared_error)

    def test_negative_log_predictive_density(self):
        self._test_metric(negative_log_predictive_density)

    def test_mean_standardized_log_loss(self):
        self._test_metric(mean_standardized_log_loss)

    def test_standardized_mean_squared_error(self):
        self._test_metric(standardized_mean_squared_error)

    def test_quantile_coverage_error(self):
        self._test_metric(
            quantile_coverage_error,
            should_check_differentiable=False,
            quantile=95.0,
        )
        # check that negative or very positive quantile coverages raise errors
        with self.assertRaises(NotImplementedError):
            self.get_metric(
                self.untrained_model,
                self.untrained_model.likelihood,
                *self.testing_pts,
                quantile_coverage_error,
                quantile=-1.0,
            )
        with self.assertRaises(NotImplementedError):
            self.get_metric(
                self.untrained_model,
                self.untrained_model.likelihood,
                *self.testing_pts,
                quantile_coverage_error,
                quantile=1000.0,
            )


class TestMetricsBatchedSingleTask(TestMetricsSingleTask):
    def setUp(self):
        batch_shape = torch.Size((4,))
        super().setUp(batch_shape=batch_shape)


class TestMetricsMultiTask(TestMetricsSingleTask):
    def create_data_and_labels(self, n=20, batch_shape=None, seed=0, is_training=True):
        input_range = [0.0, 1.0] if is_training else [1.0, 1.2]
        torch.random.manual_seed(seed)
        if batch_shape is None:
            batch_shape = torch.Size()
        inputs = torch.linspace(*input_range, n).view(-1, 1).contiguous()
        if batch_shape != torch.Size():
            inputs = inputs.unsqueeze(0)
            inputs = inputs.repeat(*batch_shape, 1, 1)
        labels = torch.cat(
            [
                torch.sin(inputs * (2 * math.pi)) + torch.randn_like(inputs) * 0.1,
                torch.cos(inputs * (2 * math.pi)) + torch.randn_like(inputs) * 0.1,
            ],
            -1,
        )
        return inputs, labels.squeeze().contiguous()

    def create_model(self, train_x, train_y):
        likelihood = MultitaskGaussianLikelihood(num_tasks=2)
        model = ExactMultiTaskGPModel(train_x, train_y, likelihood)
        return model

    def _test_metric(self, metric, should_check_differentiable=True, **kwargs):
        init_value, final_value = self.check_metric(metric, **kwargs)
        # here the trailing dimension is the number of tasks
        self.assertEqual(
            final_value.shape,
            torch.Size(
                (
                    *self.train_y.shape[:-2],
                    self.train_y.shape[-1],
                )
            ),
        )
        self.assertTrue(torch.all(final_value.mean() <= init_value.mean()))
        if should_check_differentiable:
            self.assertTrue(final_value.requires_grad)  # check that the metric is differentiable

    def test_negative_log_predictive_density(self):
        init_value, final_value = self.check_metric(negative_log_predictive_density)
        self.assertEqual(final_value.shape, self.train_y.shape[:-2])
        self.assertTrue(torch.all(final_value <= init_value))
        self.assertTrue(final_value.requires_grad)  # check that the metric is differentiable


class TestMetricsBatchedMultiTask(TestMetricsMultiTask):
    def setUp(self):
        batch_shape = torch.Size((4,))
        super().setUp(batch_shape=batch_shape)



#!/usr/bin/env python3
import math

from linear_operator.operators import KernelLinearOperator
from torch import Tensor

from .keops_kernel import _Anysor, _lazify_and_expand_inputs, KeOpsKernel


def _covar_func(x1: _Anysor, x2: _Anysor, nu: float = 2.5, **params) -> _Anysor:
    x1_, x2_ = _lazify_and_expand_inputs(x1, x2)

    sq_distance = ((x1_ - x2_) ** 2).sum(-1)
    distance = (sq_distance + 1e-20).sqrt()
    # ^^ Need to add epsilon to prevent small negative values with the sqrt
    # backward pass (otherwise we get NaNs).
    # using .clamp(1e-20, math.inf) doesn't work in KeOps; it also creates NaNs
    exp_component = (-math.sqrt(nu * 2) * distance).exp()

    if nu == 0.5:
        constant_component = 1
    elif nu == 1.5:
        constant_component = (math.sqrt(3) * distance) + 1
    elif nu == 2.5:
        constant_component = (math.sqrt(5) * distance) + (1 + 5.0 / 3.0 * sq_distance)

    return constant_component * exp_component


class MaternKernel(KeOpsKernel):
    """
    Implements the Matern kernel using KeOps as a driver for kernel matrix multiplies.

    This class can be used as a drop in replacement for :class:`gpytorch.kernels.MaternKernel` in most cases,
    and supports the same arguments.

    :param nu: (Default: 2.5) The smoothness parameter.
    :type nu: float (0.5, 1.5, or 2.5)
    :param ard_num_dims: (Default: `None`) Set this if you want a separate lengthscale for each
        input dimension. It should be `d` if x1 is a `... x n x d` matrix.
    :type ard_num_dims: int, optional
    :param batch_shape: (Default: `None`) Set this if you want a separate lengthscale for each
         batch of input data. It should be `torch.Size([b1, b2])` for a `b1 x b2 x n x m` kernel output.
    :type batch_shape: torch.Size, optional
    :param active_dims: (Default: `None`) Set this if you want to
        compute the covariance of only a few input dimensions. The ints
        corresponds to the indices of the dimensions.
    :type active_dims: Tuple(int)
    :param lengthscale_prior: (Default: `None`)
        Set this if you want to apply a prior to the lengthscale parameter.
    :type lengthscale_prior: ~gpytorch.priors.Prior, optional
    :param lengthscale_constraint: (Default: `Positive`) Set this if you want
        to apply a constraint to the lengthscale parameter.
    :type lengthscale_constraint: ~gpytorch.constraints.Interval, optional
    :param eps: (Default: 1e-6) The minimum value that the lengthscale can take (prevents divide by zero errors).
    :type eps: float, optional
    """

    has_lengthscale = True

    def __init__(self, nu: float = 2.5, **kwargs):
        if nu not in {0.5, 1.5, 2.5}:
            raise RuntimeError("nu expected to be 0.5, 1.5, or 2.5")
        super().__init__(**kwargs)
        self.nu = nu

    def forward(self, x1: Tensor, x2: Tensor, diag: bool = False, **kwargs) -> KernelLinearOperator:
        mean = x1.reshape(-1, x1.size(-1)).mean(0)[(None,) * (x1.dim() - 1)]
        x1_ = (x1 - mean) / self.lengthscale
        x2_ = (x2 - mean) / self.lengthscale
        # return KernelLinearOperator inst only when calculating the whole covariance matrix
        res = KernelLinearOperator(x1_, x2_, covar_func=_covar_func, nu=self.nu, **kwargs)

        # TODO: diag=True mode will be removed with the GpyTorch 2.0 PR to remove LazyEvaluatedKernelTensor
        # (it will be replaced by a `_symmetric_diag` method for quickly computing the diagonals of symmetric matrices)
        if diag:
            return res.diagonal(dim1=-1, dim2=-2)
        else:
            return res



#!/usr/bin/env python3

from typing import Optional, Tuple

import torch
from linear_operator import to_linear_operator
from linear_operator.operators import DiagLinearOperator, LinearOperator, MatmulLinearOperator, SumLinearOperator
from linear_operator.utils import linear_cg
from torch import Tensor
from torch.autograd.function import FunctionCtx

from .. import settings
from ..distributions import Delta, Distribution, MultivariateNormal
from ..module import Module
from ..utils.memoize import cached
from ._variational_strategy import _VariationalStrategy
from .natural_variational_distribution import NaturalVariationalDistribution


class _NgdInterpTerms(torch.autograd.Function):
    """
    This function takes in

        - the kernel interpolation term K_ZZ^{-1/2} k_ZX
        - the natural parameters of the variational distribution

    and returns

        - the predictive distribution mean/covariance
        - the inducing KL divergence KL( q(u) || p(u))

    However, the gradients will be with respect to the **cannonical parameters**
    of the variational distribution, rather than the **natural parameters**.
    This corresponds to performing natural gradient descent on the variational distribution.
    """

    @staticmethod
    def forward(
        ctx: FunctionCtx,
        interp_term: torch.Tensor,
        natural_vec: torch.Tensor,
        natural_mat: torch.Tensor,
    ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        # Compute precision
        prec = natural_mat.mul(-2.0)
        diag = prec.diagonal(dim1=-1, dim2=-2).unsqueeze(-1)

        # Make sure that interp_term and natural_vec are the same batch shape
        batch_shape = torch.broadcast_shapes(interp_term.shape[:-2], natural_vec.shape[:-1])
        expanded_interp_term = interp_term.expand(*batch_shape, *interp_term.shape[-2:])
        expanded_natural_vec = natural_vec.expand(*batch_shape, natural_vec.size(-1))

        # Compute necessary solves with the precision. We need
        # m = expec_vec = S * natural_vec
        # S K^{-1/2} k
        solves = linear_cg(
            prec.matmul,
            torch.cat([expanded_natural_vec.unsqueeze(-1), expanded_interp_term], dim=-1),
            n_tridiag=0,
            max_iter=settings.max_cg_iterations.value(),
            tolerance=min(settings.eval_cg_tolerance.value(), settings.cg_tolerance.value()),
            max_tridiag_iter=settings.max_lanczos_quadrature_iterations.value(),
            preconditioner=lambda x: x / diag,
        )
        expec_vec = solves[..., 0]
        s_times_interp_term = solves[..., 1:]

        # Compute the interpolated mean
        # k^T K^{-1/2} m
        interp_mean = (s_times_interp_term.transpose(-1, -2) @ natural_vec.unsqueeze(-1)).squeeze(-1)

        # Compute the interpolated variance
        # k^T K^{-1/2} S K^{-1/2} k = k^T K^{-1/2} (expec_mat - expec_vec expec_vec^T) K^{-1/2} k
        interp_var = (s_times_interp_term * interp_term).sum(dim=-2)

        # Let's not bother actually computing the KL-div in the foward pass
        # 1/2 ( -log | S | + tr(S) + m^T m - len(m) )
        # = 1/2 ( -log | expec_mat - expec_vec expec_vec^T | + tr(expec_mat) - len(m) )
        kl_div = torch.zeros_like(interp_mean[..., 0])

        # We're done!
        ctx.save_for_backward(interp_term, s_times_interp_term, interp_mean, natural_vec, expec_vec, prec)
        return interp_mean, interp_var, kl_div

    @staticmethod
    def backward(
        ctx: FunctionCtx, interp_mean_grad: torch.Tensor, interp_var_grad: torch.Tensor, kl_div_grad: torch.Tensor
    ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor, None]:
        # Get the saved terms
        interp_term, s_times_interp_term, interp_mean, natural_vec, expec_vec, prec = ctx.saved_tensors

        # Expand data-depenedent gradients
        interp_mean_grad = interp_mean_grad.unsqueeze(-2)
        interp_var_grad = interp_var_grad.unsqueeze(-2)

        # Compute gradient of interp term (K^{-1/2} k)
        # interp_mean component: m
        # interp_var component: S K^{-1/2} k
        # kl component: 0
        interp_term_grad = (interp_var_grad * s_times_interp_term).mul(2.0) + (
            interp_mean_grad * expec_vec.unsqueeze(-1)
        )

        # Compute gradient of expected vector (m)
        # interp_mean component: K^{-1/2} k
        # interp_var component: (k^T K^{-1/2} m) K^{-1/2} k
        # kl component: S^{-1} m
        expec_vec_grad = (
            (interp_var_grad * interp_mean.unsqueeze(-2) * interp_term).sum(dim=-1).mul(-2)
            + (interp_mean_grad * interp_term).sum(dim=-1)
            + (kl_div_grad.unsqueeze(-1) * natural_vec)
        )

        # Compute gradient of expected matrix (mm^T + S)
        # interp_mean component: 0
        # interp_var component: K^{-1/2} k k^T K^{-1/2}
        # kl component: 1/2 ( I - S^{-1} )
        eye = torch.eye(expec_vec.size(-1), device=expec_vec.device, dtype=expec_vec.dtype)
        expec_mat_grad = torch.add(
            (interp_var_grad * interp_term) @ interp_term.transpose(-1, -2),
            (kl_div_grad.unsqueeze(-1).unsqueeze(-1) * (eye - prec).mul(0.5)),
        )

        # We're done!
        return interp_term_grad, expec_vec_grad, expec_mat_grad, None  # Extra "None" for the kwarg


class CiqVariationalStrategy(_VariationalStrategy):
    r"""
    Similar to :class:`~gpytorch.variational.VariationalStrategy`,
    except the whitening operation is performed using Contour Integral Quadrature
    rather than Cholesky (see `Pleiss et al. (2020)`_ for more info).
    See the `CIQ-SVGP tutorial`_ for an example.

    Contour Integral Quadrature uses iterative matrix-vector multiplication to approximate
    the :math:`\mathbf K_{\mathbf Z \mathbf Z}^{-1/2}` matrix used for the whitening operation.
    This can be more efficient than the standard variational strategy for large numbers
    of inducing points (e.g. :math:`M > 1000`) or when the inducing points have structure
    (e.g. they lie on an evenly-spaced grid).

    .. note::

        It is recommended that this object is used in conjunction with
        :obj:`~gpytorch.variational.NaturalVariationalDistribution` and
        `natural gradient descent`_.

    :param model: Model this strategy is applied to.
        Typically passed in when the VariationalStrategy is created in the
        __init__ method of the user defined model.
    :param inducing_points: Tensor containing a set of inducing
        points to use for variational inference.
    :param variational_distribution: A
        VariationalDistribution object that represents the form of the variational distribution :math:`q(\mathbf u)`
    :param learn_inducing_locations: (Default True): Whether or not
        the inducing point locations :math:`\mathbf Z` should be learned (i.e. are they
        parameters of the model).
    :param jitter_val: Amount of diagonal jitter to add for Cholesky factorization numerical stability

    .. _Pleiss et al. (2020):
        https://arxiv.org/pdf/2006.11267.pdf
    .. _CIQ-SVGP tutorial:
        examples/04_Variational_and_Approximate_GPs/SVGP_CIQ.html
    .. _natural gradient descent:
        examples/04_Variational_and_Approximate_GPs/Natural_Gradient_Descent.html
    """

    def _ngd(self) -> bool:
        return isinstance(self._variational_distribution, NaturalVariationalDistribution)

    @property
    @cached(name="prior_distribution_memo")
    def prior_distribution(self) -> MultivariateNormal:
        zeros = torch.zeros(
            self._variational_distribution.shape(),
            dtype=self._variational_distribution.dtype,
            device=self._variational_distribution.device,
        )
        ones = torch.ones_like(zeros)
        res = MultivariateNormal(zeros, DiagLinearOperator(ones))
        return res

    @property
    @cached(name="variational_distribution_memo")
    def variational_distribution(self) -> Distribution:
        if self._ngd():
            raise RuntimeError(
                "Variational distribution for NGD-CIQ should be computed during forward calls. "
                "This is probably a bug in GPyTorch."
            )
        return super().variational_distribution

    def forward(
        self,
        x: torch.Tensor,
        inducing_points: torch.Tensor,
        inducing_values: torch.Tensor,
        variational_inducing_covar: Optional[LinearOperator] = None,
        *params,
        **kwargs,
    ) -> MultivariateNormal:
        # Compute full prior distribution
        full_inputs = torch.cat([inducing_points, x], dim=-2)
        full_output = self.model.forward(full_inputs, *params, **kwargs)
        full_covar = full_output.lazy_covariance_matrix

        # Covariance terms
        num_induc = inducing_points.size(-2)
        test_mean = full_output.mean[..., num_induc:]
        induc_induc_covar = full_covar[..., :num_induc, :num_induc].evaluate_kernel().add_jitter(self.jitter_val)
        induc_data_covar = full_covar[..., :num_induc, num_induc:].to_dense()
        data_data_covar = full_covar[..., num_induc:, num_induc:].add_jitter(self.jitter_val)

        # Compute interpolation terms
        # K_XZ K_ZZ^{-1} \mu_z
        # K_XZ K_ZZ^{-1/2} \mu_Z
        with settings.max_preconditioner_size(0):  # Turn off preconditioning for CIQ
            interp_term = to_linear_operator(induc_induc_covar).sqrt_inv_matmul(induc_data_covar)

        # Compute interpolated mean and variance terms
        # We have separate computation rules for NGD versus standard GD
        if self._ngd():
            interp_mean, interp_var, kl_div = _NgdInterpTerms.apply(
                interp_term,
                self._variational_distribution.natural_vec,
                self._variational_distribution.natural_mat,
            )

            # Compute the covariance of q(f)
            predictive_var = data_data_covar.diagonal(dim1=-1, dim2=-2) - interp_term.pow(2).sum(dim=-2) + interp_var
            predictive_var = torch.clamp_min(predictive_var, settings.min_variance.value(predictive_var.dtype))
            predictive_covar = DiagLinearOperator(predictive_var)

            # Also compute and cache the KL divergence
            if not hasattr(self, "_memoize_cache"):
                self._memoize_cache = dict()
            self._memoize_cache["kl"] = kl_div

        else:
            # Compute interpolated mean term
            interp_mean = torch.matmul(
                interp_term.transpose(-1, -2), (inducing_values - self.prior_distribution.mean).unsqueeze(-1)
            ).squeeze(-1)

            # Compute the covariance of q(f)
            middle_term = self.prior_distribution.lazy_covariance_matrix.mul(-1)
            if variational_inducing_covar is not None:
                middle_term = SumLinearOperator(variational_inducing_covar, middle_term)
            predictive_covar = SumLinearOperator(
                data_data_covar.add_jitter(self.jitter_val),
                MatmulLinearOperator(interp_term.transpose(-1, -2), middle_term @ interp_term),
            )

        # Compute the mean of q(f)
        # k_XZ K_ZZ^{-1/2} (m - K_ZZ^{-1/2} \mu_Z) + \mu_X
        predictive_mean = interp_mean + test_mean

        # Return the distribution
        return MultivariateNormal(predictive_mean, predictive_covar)

    def kl_divergence(self) -> Tensor:
        r"""
        Compute the KL divergence between the variational inducing distribution :math:`q(\mathbf u)`
        and the prior inducing distribution :math:`p(\mathbf u)`.

        :rtype: torch.Tensor
        """
        if self._ngd():
            if hasattr(self, "_memoize_cache") and "kl" in self._memoize_cache:
                return self._memoize_cache["kl"]
            else:
                raise RuntimeError(
                    "KL divergence for NGD-CIQ should be computed during forward calls."
                    "This is probably a bug in GPyTorch."
                )
        else:
            return super().kl_divergence()

    def __call__(self, x: torch.Tensor, prior: bool = False, *params, **kwargs) -> MultivariateNormal:
        # This is mostly the same as _VariationalStrategy.__call__()
        # but with special rules for natural gradient descent (to prevent O(M^3) computation)

        # If we're in prior mode, then we're done!
        if prior:
            return self.model.forward(x)

        # Delete previously cached items from the training distribution
        if self.training:
            self._clear_cache()

        # (Maybe) initialize variational distribution
        if not self.variational_params_initialized.item():
            if self._ngd():
                noise = torch.randn_like(self.prior_distribution.mean).mul_(1e-3)
                eye = torch.eye(noise.size(-1), dtype=noise.dtype, device=noise.device).mul(-0.5)
                self._variational_distribution.natural_vec.data.copy_(noise)
                self._variational_distribution.natural_mat.data.copy_(eye)
                self.variational_params_initialized.fill_(1)
            else:
                prior_dist = self.prior_distribution
                self._variational_distribution.initialize_variational_distribution(prior_dist)
                self.variational_params_initialized.fill_(1)

        # Ensure inducing_points and x are the same size
        inducing_points = self.inducing_points
        if inducing_points.shape[:-2] != x.shape[:-2]:
            x, inducing_points = self._expand_inputs(x, inducing_points)

        # Get q(f)
        if self._ngd():
            return Module.__call__(
                self,
                x,
                inducing_points,
                inducing_values=None,
                variational_inducing_covar=None,
                *params,
                **kwargs,
            )
        else:
            # Get p(u)/q(u)
            variational_dist_u = self.variational_distribution

            if isinstance(variational_dist_u, MultivariateNormal):
                return Module.__call__(
                    self,
                    x,
                    inducing_points,
                    inducing_values=variational_dist_u.mean,
                    variational_inducing_covar=variational_dist_u.lazy_covariance_matrix,
                    **kwargs,
                )
            elif isinstance(variational_dist_u, Delta):
                return Module.__call__(
                    self,
                    x,
                    inducing_points,
                    inducing_values=variational_dist_u.mean,
                    variational_inducing_covar=None,
                    **kwargs,
                )
            else:
                raise RuntimeError(
                    f"Invalid variational distribuition ({type(variational_dist_u)}). "
                    "Expected a multivariate normal or a delta distribution."
                )



#!/usr/bin/env python3

# This implementation is based on the DenseNet-BC implementation in torchvision
# https://github.com/pytorch/vision/blob/master/torchvision/models/densenet.py

from collections import OrderedDict

import torch
import torch.nn as nn
import torch.nn.functional as F


class _DenseLayer(nn.Sequential):
    def __init__(self, num_input_features, growth_rate, bn_size, drop_rate):
        super(_DenseLayer, self).__init__()
        self.add_module("norm1", nn.BatchNorm2d(num_input_features)),
        self.add_module("relu1", nn.ReLU(inplace=True)),
        self.add_module(
            "conv1", nn.Conv2d(num_input_features, bn_size * growth_rate, kernel_size=1, stride=1, bias=False)
        ),
        self.add_module("norm2", nn.BatchNorm2d(bn_size * growth_rate)),
        self.add_module("relu2", nn.ReLU(inplace=True)),
        self.add_module(
            "conv2", nn.Conv2d(bn_size * growth_rate, growth_rate, kernel_size=3, stride=1, padding=1, bias=False)
        ),
        self.drop_rate = drop_rate

    def forward(self, x):
        new_features = super(_DenseLayer, self).forward(x)
        if self.drop_rate > 0:
            new_features = F.dropout(new_features, p=self.drop_rate, training=self.training)
        return torch.cat([x, new_features], 1)


class _Transition(nn.Sequential):
    def __init__(self, num_input_features, num_output_features):
        super(_Transition, self).__init__()
        self.add_module("norm", nn.BatchNorm2d(num_input_features))
        self.add_module("relu", nn.ReLU(inplace=True))
        self.add_module("conv", nn.Conv2d(num_input_features, num_output_features, kernel_size=1, stride=1, bias=False))
        self.add_module("pool", nn.AvgPool2d(kernel_size=2, stride=2))


class _DenseBlock(nn.Sequential):
    def __init__(self, num_layers, num_input_features, bn_size, growth_rate, drop_rate):
        super(_DenseBlock, self).__init__()
        for i in range(num_layers):
            layer = _DenseLayer(num_input_features + i * growth_rate, growth_rate, bn_size, drop_rate)
            self.add_module("denselayer%d" % (i + 1), layer)


class DenseNet(nn.Module):
    r"""Densenet-BC model class, based on
    `"Densely Connected Convolutional Networks" <https://arxiv.org/pdf/1608.06993.pdf>`
    Args:
        growth_rate (int) - how many filters to add each layer (`k` in paper)
        block_config (list of 3 or 4 ints) - how many layers in each pooling block
        num_init_features (int) - the number of filters to learn in the first convolution layer
        bn_size (int) - multiplicative factor for number of bottle neck layers
            (i.e. bn_size * k features in the bottleneck layer)
        drop_rate (float) - dropout rate after each dense layer
        num_classes (int) - number of classification classes
    """

    def __init__(
        self,
        growth_rate=12,
        block_config=(16, 16, 16),
        compression=0.5,
        num_init_features=24,
        bn_size=4,
        drop_rate=0,
        avgpool_size=8,
        num_classes=10,
    ):

        super(DenseNet, self).__init__()
        assert 0 < compression <= 1, "compression of densenet should be between 0 and 1"
        self.avgpool_size = avgpool_size

        # First convolution
        self.features = nn.Sequential(
            OrderedDict([("conv0", nn.Conv2d(3, num_init_features, kernel_size=3, stride=1, padding=1, bias=False))])
        )

        # Each denseblock
        num_features = num_init_features
        for i, num_layers in enumerate(block_config):
            block = _DenseBlock(
                num_layers=num_layers,
                num_input_features=num_features,
                bn_size=bn_size,
                growth_rate=growth_rate,
                drop_rate=drop_rate,
            )
            self.features.add_module("denseblock%d" % (i + 1), block)
            num_features = num_features + num_layers * growth_rate
            if i != len(block_config) - 1:
                trans = _Transition(
                    num_input_features=num_features, num_output_features=int(num_features * compression)
                )
                self.features.add_module("transition%d" % (i + 1), trans)
                num_features = int(num_features * compression)

        # Final batch norm
        self.features.add_module("norm_final", nn.BatchNorm2d(num_features))

        # Linear layer
        self.classifier = nn.Linear(num_features, num_classes)

    def forward(self, x):
        features = self.features(x)
        out = F.relu(features, inplace=True)
        out = F.avg_pool2d(out, kernel_size=self.avgpool_size).view(features.size(0), -1)
        out = self.classifier(out)
        return out



#!/usr/bin/env python3

from ..functions import RBFCovariance
from ..settings import trace_mode
from .kernel import Kernel


def postprocess_rbf(dist_mat):
    return dist_mat.div_(-2).exp_()


class RBFKernel(Kernel):
    r"""
    Computes a covariance matrix based on the RBF (squared exponential) kernel
    between inputs :math:`\mathbf{x_1}` and :math:`\mathbf{x_2}`:

    .. math::

       \begin{equation*}
          k_{\text{RBF}}(\mathbf{x_1}, \mathbf{x_2}) = \exp \left( -\frac{1}{2}
          (\mathbf{x_1} - \mathbf{x_2})^\top \Theta^{-2} (\mathbf{x_1} - \mathbf{x_2}) \right)
       \end{equation*}

    where :math:`\Theta` is a lengthscale parameter.
    See :class:`gpytorch.kernels.Kernel` for descriptions of the lengthscale options.

    .. note::

        This kernel does not have an `outputscale` parameter. To add a scaling parameter,
        decorate this kernel with a :class:`gpytorch.kernels.ScaleKernel`.

    :param ard_num_dims: Set this if you want a separate lengthscale for each input
        dimension. It should be `d` if :math:`\mathbf{x_1}` is a `n x d` matrix. (Default: `None`.)
    :param batch_shape: Set this if you want a separate lengthscale for each batch of input
        data. It should be :math:`B_1 \times \ldots \times B_k` if :math:`\mathbf{x_1}` is
        a :math:`B_1 \times \ldots \times B_k \times N \times D` tensor.
    :param active_dims: Set this if you want to compute the covariance of only
        a few input dimensions. The ints corresponds to the indices of the
        dimensions. (Default: `None`.)
    :param lengthscale_prior: Set this if you want to apply a prior to the
        lengthscale parameter. (Default: `None`)
    :param lengthscale_constraint: Set this if you want to apply a constraint
        to the lengthscale parameter. (Default: `Positive`.)
    :param eps: The minimum value that the lengthscale can take (prevents
        divide by zero errors). (Default: `1e-6`.)

    :ivar torch.Tensor lengthscale: The lengthscale parameter. Size/shape of parameter depends on the
        ard_num_dims and batch_shape arguments.

    Example:
        >>> x = torch.randn(10, 5)
        >>> # Non-batch: Simple option
        >>> covar_module = gpytorch.kernels.ScaleKernel(gpytorch.kernels.RBFKernel())
        >>> # Non-batch: ARD (different lengthscale for each input dimension)
        >>> covar_module = gpytorch.kernels.ScaleKernel(gpytorch.kernels.RBFKernel(ard_num_dims=5))
        >>> covar = covar_module(x)  # Output: LinearOperator of size (10 x 10)
        >>>
        >>> batch_x = torch.randn(2, 10, 5)
        >>> # Batch: Simple option
        >>> covar_module = gpytorch.kernels.ScaleKernel(gpytorch.kernels.RBFKernel())
        >>> # Batch: different lengthscale for each batch
        >>> covar_module = gpytorch.kernels.ScaleKernel(gpytorch.kernels.RBFKernel(batch_shape=torch.Size([2])))
        >>> covar = covar_module(x)  # Output: LinearOperator of size (2 x 10 x 10)
    """

    has_lengthscale = True

    def forward(self, x1, x2, diag=False, **params):
        if (
            x1.requires_grad
            or x2.requires_grad
            or (self.ard_num_dims is not None and self.ard_num_dims > 1)
            or diag
            or params.get("last_dim_is_batch", False)
            or trace_mode.on()
        ):
            x1_ = x1.div(self.lengthscale)
            x2_ = x2.div(self.lengthscale)
            return postprocess_rbf(self.covar_dist(x1_, x2_, square_dist=True, diag=diag, **params))
        return RBFCovariance.apply(
            x1,
            x2,
            self.lengthscale,
            lambda x1, x2: self.covar_dist(x1, x2, square_dist=True, diag=False, **params),
        )



#!/usr/bin/env python3

import math
import warnings
from abc import ABC, abstractmethod
from copy import deepcopy
from typing import Any, Dict, Optional, Union

import torch
from torch import Tensor
from torch.distributions import Distribution as _Distribution

from .. import settings
from ..distributions import base_distributions, MultivariateNormal
from ..module import Module
from ..utils.quadrature import GaussHermiteQuadrature1D
from ..utils.warnings import GPInputWarning


class _Likelihood(Module, ABC):
    has_analytic_marginal: bool = False

    def __init__(self, max_plate_nesting: int = 1) -> None:
        super().__init__()
        self.max_plate_nesting: int = max_plate_nesting

    def _draw_likelihood_samples(
        self, function_dist: MultivariateNormal, *args: Any, sample_shape: Optional[torch.Size] = None, **kwargs: Any
    ) -> _Distribution:
        if sample_shape is None:
            sample_shape = torch.Size(
                [settings.num_likelihood_samples.value()]
                + [1] * (self.max_plate_nesting - len(function_dist.batch_shape) - 1)
            )
        else:
            sample_shape = sample_shape[: -len(function_dist.batch_shape) - 1]
        if self.training:
            num_event_dims = len(function_dist.event_shape)
            function_dist = base_distributions.Normal(function_dist.mean, function_dist.variance.sqrt())
            function_dist = base_distributions.Independent(function_dist, num_event_dims - 1)
        function_samples = function_dist.rsample(sample_shape)
        return self.forward(function_samples, *args, **kwargs)

    def expected_log_prob(
        self, observations: Tensor, function_dist: MultivariateNormal, *args: Any, **kwargs: Any
    ) -> Tensor:
        likelihood_samples = self._draw_likelihood_samples(function_dist, *args, **kwargs)
        res = likelihood_samples.log_prob(observations, *args, **kwargs).mean(dim=0)
        return res

    @abstractmethod
    def forward(self, function_samples: Tensor, *args: Any, **kwargs: Any) -> _Distribution:
        raise NotImplementedError

    def get_fantasy_likelihood(self, **kwargs: Any) -> "_Likelihood":
        return deepcopy(self)

    def log_marginal(
        self, observations: Tensor, function_dist: MultivariateNormal, *args: Any, **kwargs: Any
    ) -> Tensor:
        likelihood_samples = self._draw_likelihood_samples(function_dist, *args, **kwargs)
        log_probs = likelihood_samples.log_prob(observations)
        res = log_probs.sub(math.log(log_probs.size(0))).logsumexp(dim=0)
        return res

    def marginal(self, function_dist: MultivariateNormal, *args: Any, **kwargs: Any) -> _Distribution:
        res = self._draw_likelihood_samples(function_dist, *args, **kwargs)
        return res

    def __call__(self, input: Union[Tensor, MultivariateNormal], *args: Any, **kwargs: Any) -> _Distribution:
        # Conditional
        if torch.is_tensor(input):
            return super().__call__(input, *args, **kwargs)  # pyre-ignore[7]
        # Marginal
        elif isinstance(input, MultivariateNormal):
            return self.marginal(input, *args, **kwargs)
        # Error
        else:
            raise RuntimeError(
                "Likelihoods expects a MultivariateNormal input to make marginal predictions, or a "
                "torch.Tensor for conditional predictions. Got a {}".format(input.__class__.__name__)
            )


try:
    import pyro

    class Likelihood(_Likelihood):
        r"""
        A Likelihood in GPyTorch specifies the mapping from latent function values
        :math:`f(\mathbf X)` to observed labels :math:`y`.

        For example, in the case of regression this might be a Gaussian
        distribution, as :math:`y(\mathbf x)` is equal to :math:`f(\mathbf x)` plus Gaussian noise:

        .. math::
            y(\mathbf x) = f(\mathbf x) + \epsilon, \:\:\:\: \epsilon \sim N(0,\sigma^{2}_{n} \mathbf I)

        In the case of classification, this might be a Bernoulli distribution,
        where the probability that :math:`y=1` is given by the latent function
        passed through some sigmoid or probit function:

        .. math::
            y(\mathbf x) = \begin{cases}
                1 & \text{w/ probability} \:\: \sigma(f(\mathbf x)) \\
                0 & \text{w/ probability} \:\: 1-\sigma(f(\mathbf x))
            \end{cases}

        In either case, to implement a likelihood function, GPyTorch only
        requires a forward method that computes the conditional distribution
        :math:`p(y \mid f(\mathbf x))`.

        :param bool has_analytic_marginal: Whether or not the marginal distribution :math:`p(\mathbf y)`
            can be computed in closed form. (See :meth:`~gpytorch.likelihoods.Likelihood.__call__` docstring.)
        :param max_plate_nesting: (For Pyro integration only.) How many batch dimensions are in the function.
            This should be modified if the likelihood uses plated random variables. (Default = 1)
            This should be modified if the likelihood uses plated random variables. (Default = 1)
        :param str name_prefix: (For Pyro integration only.) Prefix to assign to named Pyro latent variables.
        :param int num_data: (For Pyro integration only.) Total amount of observations.
        """

        @property
        def num_data(self) -> int:
            if hasattr(self, "_num_data"):
                return self._num_data
            else:
                warnings.warn(
                    "likelihood.num_data isn't set. This might result in incorrect ELBO scaling.", GPInputWarning
                )
                return ""

        @num_data.setter
        def num_data(self, val: int) -> None:
            self._num_data = val

        @property
        def name_prefix(self) -> str:
            if hasattr(self, "_name_prefix"):
                return self._name_prefix
            else:
                return ""

        @name_prefix.setter
        def name_prefix(self, val: str) -> None:
            self._name_prefix = val

        def _draw_likelihood_samples(
            self, function_dist: _Distribution, *args: Any, sample_shape: Optional[torch.Size] = None, **kwargs: Any
        ) -> _Distribution:
            if self.training:
                num_event_dims = len(function_dist.event_shape)
                function_dist = base_distributions.Normal(function_dist.mean, function_dist.variance.sqrt())
                function_dist = base_distributions.Independent(function_dist, num_event_dims - 1)

            plate_name = self.name_prefix + ".num_particles_vectorized"
            num_samples = settings.num_likelihood_samples.value()
            max_plate_nesting = max(self.max_plate_nesting, len(function_dist.batch_shape))
            with pyro.plate(plate_name, size=num_samples, dim=(-max_plate_nesting - 1)):
                if sample_shape is None:
                    function_samples = pyro.sample(self.name_prefix, function_dist.mask(False))
                    # Deal with the fact that we're not assuming conditional independence over data points here
                    function_samples = function_samples.squeeze(-len(function_dist.event_shape) - 1)
                else:
                    sample_shape = sample_shape[: -len(function_dist.batch_shape)]
                    function_samples = function_dist(sample_shape)

                if not self.training:
                    function_samples = function_samples.squeeze(-len(function_dist.event_shape) - 1)
                return self.forward(function_samples, *args, **kwargs)

        def expected_log_prob(
            self, observations: Tensor, function_dist: MultivariateNormal, *args: Any, **kwargs: Any
        ) -> Tensor:
            r"""
            (Used by :obj:`~gpytorch.mlls.VariationalELBO` for variational inference.)

            Computes the expected log likelihood, where the expectation is over the GP variational distribution.

            .. math::
                \sum_{\mathbf x, y} \mathbb{E}_{q\left( f(\mathbf x) \right)}
                \left[ \log p \left( y \mid f(\mathbf x) \right) \right]

            :param observations: Values of :math:`y`.
            :param function_dist: Distribution for :math:`f(x)`.
            :param args: Additional args (passed to the forward function).
            :param kwargs: Additional kwargs (passed to the forward function).
            """
            return super().expected_log_prob(observations, function_dist, *args, **kwargs)

        @abstractmethod
        def forward(
            self, function_samples: Tensor, *args: Any, data: Dict[str, Tensor] = {}, **kwargs: Any
        ) -> _Distribution:
            r"""
            Computes the conditional distribution :math:`p(\mathbf y \mid
            \mathbf f, \ldots)` that defines the likelihood.

            :param function_samples: Samples from the function (:math:`\mathbf f`)
            :param data: (Pyro integration only.) Additional variables that the likelihood needs to condition
                on. The keys of the dictionary will correspond to Pyro sample sites
                in the likelihood's model/guide.
            :param args: Additional args
            :param kwargs: Additional kwargs
            """
            raise NotImplementedError

        def get_fantasy_likelihood(self, **kwargs: Any) -> "_Likelihood":
            """"""
            return super().get_fantasy_likelihood(**kwargs)

        def log_marginal(
            self, observations: Tensor, function_dist: MultivariateNormal, *args: Any, **kwargs: Any
        ) -> Tensor:
            r"""
            (Used by :obj:`~gpytorch.mlls.PredictiveLogLikelihood` for approximate inference.)

            Computes the log marginal likelihood of the approximate predictive distribution

            .. math::
                \sum_{\mathbf x, y} \log \mathbb{E}_{q\left( f(\mathbf x) \right)}
                \left[ p \left( y \mid f(\mathbf x) \right) \right]

            Note that this differs from :meth:`expected_log_prob` because the :math:`log` is on the outside
            of the expectation.

            :param observations: Values of :math:`y`.
            :param function_dist: Distribution for :math:`f(x)`.
            :param args: Additional args (passed to the forward function).
            :param kwargs: Additional kwargs (passed to the forward function).
            """
            return super().log_marginal(observations, function_dist, *args, **kwargs)

        def marginal(self, function_dist: MultivariateNormal, *args: Any, **kwargs: Any) -> _Distribution:
            r"""
            Computes a predictive distribution :math:`p(y^* | \mathbf x^*)` given either a posterior
            distribution :math:`p(\mathbf f | \mathcal D, \mathbf x)` or a
            prior distribution :math:`p(\mathbf f|\mathbf x)` as input.

            With both exact inference and variational inference, the form of
            :math:`p(\mathbf f|\mathcal D, \mathbf x)` or :math:`p(\mathbf f|
            \mathbf x)` should usually be Gaussian. As a result, function_dist
            should usually be a :obj:`~gpytorch.distributions.MultivariateNormal` specified by the mean and
            (co)variance of :math:`p(\mathbf f|...)`.

            :param function_dist: Distribution for :math:`f(x)`.
            :param args: Additional args (passed to the forward function).
            :param kwargs: Additional kwargs (passed to the forward function).
            :return: The marginal distribution, or samples from it.
            """
            return super().marginal(function_dist, *args, **kwargs)

        def pyro_guide(self, function_dist: MultivariateNormal, target: Tensor, *args: Any, **kwargs: Any) -> None:
            r"""
            (For Pyro integration only).

            Part of the guide function for the likelihood.
            This should be re-defined if the likelihood contains any latent variables that need to be infered.

            :param function_dist: Distribution of latent function
                :math:`q(\mathbf f)`.
            :param target: Observed :math:`\mathbf y`.
            :param args: Additional args (passed to the forward function).
            :param kwargs: Additional kwargs (passed to the forward function).
            """
            with pyro.plate(self.name_prefix + ".data_plate", dim=-1):
                pyro.sample(self.name_prefix + ".f", function_dist)

        def pyro_model(self, function_dist: MultivariateNormal, target: Tensor, *args: Any, **kwargs: Any) -> Tensor:
            r"""
            (For Pyro integration only).

            Part of the model function for the likelihood.
            It should return the
            This should be re-defined if the likelihood contains any latent variables that need to be infered.

            :param function_dist: Distribution of latent function
                :math:`p(\mathbf f)`.
            :param target: Observed :math:`\mathbf y`.
            :param args: Additional args (passed to the forward function).
            :param kwargs: Additional kwargs (passed to the forward function).
            """
            with pyro.plate(self.name_prefix + ".data_plate", dim=-1):
                function_samples = pyro.sample(self.name_prefix + ".f", function_dist)
                output_dist = self(function_samples, *args, **kwargs)
                return self.sample_target(output_dist, target)

        def sample_target(self, output_dist: MultivariateNormal, target: Tensor) -> Tensor:
            scale = (self.num_data or output_dist.batch_shape[-1]) / output_dist.batch_shape[-1]
            with pyro.poutine.scale(scale=scale):  # pyre-ignore[16]
                return pyro.sample(self.name_prefix + ".y", output_dist, obs=target)

        def __call__(self, input: Union[Tensor, MultivariateNormal], *args: Any, **kwargs: Any) -> _Distribution:
            r"""
            Calling this object does one of two things:

            1. If likelihood is called with a :class:`torch.Tensor` object, then it is
               assumed that the input is samples from :math:`f(\mathbf x)`. This
               returns the *conditional* distribution :math:`p(y|f(\mathbf x))`.

            .. code-block:: python

                f = torch.randn(20)
                likelihood = gpytorch.likelihoods.GaussianLikelihood()
                conditional = likelihood(f)
                print(type(conditional), conditional.batch_shape, conditional.event_shape)
                # >>> torch.distributions.Normal, torch.Size([20]), torch.Size([])

            2. If likelihood is called with a :class:`~gpytorch.distribution.MultivariateNormal` object,
               then it is assumed that the input is the distribution :math:`f(\mathbf x)`.
               This returns the *marginal* distribution :math:`p(y|\mathbf x)`.

               The form of the marginal distribution depends on the likelihood.
               For :class:`~gpytorch.likelihoods.BernoulliLikelihood` and
               :class:`~gpytorch.likelihoods.GaussianLikelihood` objects, the marginal distribution
               can be computed analytically, and the likelihood returns the analytic distribution.
               For most other likelihoods, there is no analytic form for the marginal,
               and so the likelihood instead returns a batch of Monte Carlo samples from the marginal.

            .. code-block:: python

                mean = torch.randn(20)
                covar = linear_operator.operators.DiagLinearOperator(torch.ones(20))
                f = gpytorch.distributions.MultivariateNormal(mean, covar)

                # Analytic marginal computation - Bernoulli and Gaussian likelihoods only
                analytic_marginal_likelihood = gpytorch.likelihoods.GaussianLikelihood()
                marginal = analytic_marginal_likelihood(f)
                print(type(marginal), marginal.batch_shape, marginal.event_shape)
                # >>> <class 'gpytorch.distributions.multivariate_normal.MultivariateNormal'> torch.Size([]) torch.Size([20])  # noqa: E501

                # MC marginal computation - all other likelihoods
                mc_marginal_likelihood = gpytorch.likelihoods.BetaLikelihood()
                with gpytorch.settings.num_likelihood_samples(15):
                    marginal = mc_marginal_likelihood(f)
                print(type(marginal), marginal.batch_shape, marginal.event_shape)
                # >>> <class 'torch.distributions.beta.Beta'> torch.Size([15, 20]) torch.Size([])
                # The batch_shape torch.Size([15, 20]) represents 15 MC samples for 20 data points.

            .. note::

                If a Likelihood supports analytic marginals, the :attr:`has_analytic_marginal` property will be True.
                If a Likelihood does not support analytic marginals, you can set the number of Monte Carlo
                samples using the :class:`gpytorch.settings.num_likelihood_samples` context manager.

            :param input: Either a (... x N) sample from :math:`\mathbf f`
                or a (... x N) MVN distribution of :math:`\mathbf f`.
            :param args: Additional args (passed to the forward function).
            :param kwargs: Additional kwargs (passed to the forward function).
            :return: Either a conditional :math:`p(\mathbf y \mid \mathbf f)`
                or marginal :math:`p(\mathbf y)`
                based on whether :attr:`input` is a Tensor or a MultivariateNormal (see above).
            """
            # Conditional
            if torch.is_tensor(input):
                return super().__call__(input, *args, **kwargs)
            # Marginal
            elif any(
                [
                    isinstance(input, MultivariateNormal),
                    isinstance(input, pyro.distributions.Normal),  # pyre-ignore[16]
                    (
                        isinstance(input, pyro.distributions.Independent)  # pyre-ignore[16]
                        and isinstance(input.base_dist, pyro.distributions.Normal)  # pyre-ignore[16]
                    ),
                ]
            ):
                return self.marginal(input, *args, **kwargs)  # pyre-ignore[6]
            # Error
            else:
                raise RuntimeError(
                    "Likelihoods expects a MultivariateNormal or Normal input to make marginal predictions, or a "
                    "torch.Tensor for conditional predictions. Got a {}".format(input.__class__.__name__)
                )

except ImportError:

    class Likelihood(_Likelihood):
        @property
        def num_data(self) -> int:
            warnings.warn("num_data is only used for likelihoods that are integrated with Pyro.", RuntimeWarning)
            return 0

        @num_data.setter
        def num_data(self, val: int) -> None:
            warnings.warn("num_data is only used for likelihoods that are integrated with Pyro.", RuntimeWarning)

        @property
        def name_prefix(self) -> str:
            warnings.warn("name_prefix is only used for likelihoods that are integrated with Pyro.", RuntimeWarning)
            return ""

        @name_prefix.setter
        def name_prefix(self, val: str) -> None:
            warnings.warn("name_prefix is only used for likelihoods that are integrated with Pyro.", RuntimeWarning)


class _OneDimensionalLikelihood(Likelihood, ABC):
    r"""
    A specific case of :obj:`~gpytorch.likelihoods.Likelihood` when the GP represents a one-dimensional
    output. (I.e. for a specific :math:`\mathbf x`, :math:`f(\mathbf x) \in \mathbb{R}`.)

    Inheriting from this likelihood reduces the variance when computing approximate GP objective functions
    by using 1D Gauss-Hermite quadrature.
    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.quadrature = GaussHermiteQuadrature1D()

    def expected_log_prob(
        self, observations: Tensor, function_dist: MultivariateNormal, *args: Any, **kwargs: Any
    ) -> Tensor:
        log_prob_lambda = lambda function_samples: self.forward(function_samples, *args, **kwargs).log_prob(
            observations
        )
        log_prob = self.quadrature(log_prob_lambda, function_dist)
        return log_prob

    def log_marginal(
        self, observations: Tensor, function_dist: MultivariateNormal, *args: Any, **kwargs: Any
    ) -> Tensor:
        prob_lambda = lambda function_samples: self.forward(function_samples).log_prob(observations).exp()
        prob = self.quadrature(prob_lambda, function_dist)
        return prob.log()



#!/usr/bin/env python3

from torch.distributions import kl_divergence

from ..distributions import MultivariateNormal
from .added_loss_term import AddedLossTerm


class KLGaussianAddedLossTerm(AddedLossTerm):
    r"""
    This class is used by variational GPLVM models.
    It adds the KL divergence between two multivariate Gaussian distributions:
    scaled by the size of the data and the number of output dimensions.

    .. math::

        D_\text{KL} \left( q(\mathbf x) \Vert p(\mathbf x) \right)


    :param q_x: The MVN distribution :math:`q(\mathbf x)`.
    :param p_x: The MVN distribution :math:`p(\mathbf x)`.
    :param n: Size of the latent space.
    :param data_dim: Dimensionality of the :math:`\mathbf Y` values.
    """

    def __init__(self, q_x: MultivariateNormal, p_x: MultivariateNormal, n: int, data_dim: int):
        super().__init__()
        self.q_x = q_x
        self.p_x = p_x
        self.n = n
        self.data_dim = data_dim

    def loss(self):
        kl_per_latent_dim = kl_divergence(self.q_x, self.p_x).sum(axis=0)  # vector of size latent_dim
        kl_per_point = kl_per_latent_dim.sum() / self.n  # scalar
        # inside the forward method of variational ELBO,
        # the added loss terms are expanded (using add_) to take the same
        # shape as the log_lik term (has shape data_dim)
        # so they can be added together. Hence, we divide by data_dim to avoid
        # overcounting the kl term
        return kl_per_point / self.data_dim



#!/usr/bin/env python3

from typing import Optional, Union

import torch
from torch import Tensor

from .. import settings
from ..constraints import Interval, Positive
from ..priors import Prior
from .kernel import Kernel


class CylindricalKernel(Kernel):
    r"""
    Computes a covariance matrix based on the Cylindrical Kernel between
    inputs :math:`\mathbf{x_1}` and :math:`\mathbf{x_2}`.
    It was proposed in `BOCK: Bayesian Optimization with Cylindrical Kernels`.
    See http://proceedings.mlr.press/v80/oh18a.html for more details

    .. note::
        The data must lie completely within the unit ball.

    Args:
        num_angular_weights (int):
            The number of components in the angular kernel
        radial_base_kernel (gpytorch.kernel):
            The base kernel for computing the radial kernel
        batch_size (int, optional):
            Set this if the data is batch of input data.
            It should be `b` if x1 is a `b x n x d` tensor. Default: `1`
        eps (float):
            Small floating point number used to improve numerical stability
            in kernel computations. Default: `1e-6`
        param_transform (function, optional):
            Set this if you want to use something other than softplus to ensure positiveness of parameters.
        inv_param_transform (function, optional):
            Set this to allow setting parameters directly in transformed space and sampling from priors.
            Automatically inferred for common transformations such as torch.exp or torch.nn.functional.softplus.
    """

    def __init__(
        self,
        num_angular_weights: int,
        radial_base_kernel: Kernel,
        eps: Optional[float] = 1e-6,
        angular_weights_prior: Optional[Prior] = None,
        angular_weights_constraint: Optional[Interval] = None,
        alpha_prior: Optional[Prior] = None,
        alpha_constraint: Optional[Interval] = None,
        beta_prior: Optional[Prior] = None,
        beta_constraint: Optional[Interval] = None,
        **kwargs,
    ):
        if angular_weights_constraint is None:
            angular_weights_constraint = Positive()

        if alpha_constraint is None:
            alpha_constraint = Positive()

        if beta_constraint is None:
            beta_constraint = Positive()

        super().__init__(**kwargs)
        self.num_angular_weights = num_angular_weights
        self.radial_base_kernel = radial_base_kernel
        self.eps = eps

        self.register_parameter(
            name="raw_angular_weights",
            parameter=torch.nn.Parameter(torch.zeros(*self.batch_shape, num_angular_weights)),
        )
        self.register_constraint("raw_angular_weights", angular_weights_constraint)
        self.register_parameter(name="raw_alpha", parameter=torch.nn.Parameter(torch.zeros(*self.batch_shape, 1)))
        self.register_constraint("raw_alpha", alpha_constraint)
        self.register_parameter(name="raw_beta", parameter=torch.nn.Parameter(torch.zeros(*self.batch_shape, 1)))
        self.register_constraint("raw_beta", beta_constraint)

        if angular_weights_prior is not None:
            if not isinstance(angular_weights_prior, Prior):
                raise TypeError("Expected gpytorch.priors.Prior but got " + type(angular_weights_prior).__name__)
            self.register_prior(
                "angular_weights_prior",
                angular_weights_prior,
                lambda m: m.angular_weights,
                lambda m, v: m._set_angular_weights(v),
            )
        if alpha_prior is not None:
            if not isinstance(alpha_prior, Prior):
                raise TypeError("Expected gpytorch.priors.Prior but got " + type(alpha_prior).__name__)
            self.register_prior("alpha_prior", alpha_prior, lambda m: m.alpha, lambda m, v: m._set_alpha(v))
        if beta_prior is not None:
            if not isinstance(beta_prior, Prior):
                raise TypeError("Expected gpytorch.priors.Prior but got " + type(beta_prior).__name__)
            self.register_prior("beta_prior", beta_prior, lambda m: m.beta, lambda m, v: m._set_beta(v))

    @property
    def angular_weights(self) -> Tensor:
        return self.raw_angular_weights_constraint.transform(self.raw_angular_weights)

    @angular_weights.setter
    def angular_weights(self, value: Tensor) -> None:
        if not torch.is_tensor(value):
            value = torch.tensor(value)

        self.initialize(raw_angular_weights=self.raw_angular_weights_constraint.inverse_transform(value))

    @property
    def alpha(self) -> Tensor:
        return self.raw_alpha_constraint.transform(self.raw_alpha)

    @alpha.setter
    def alpha(self, value: Tensor) -> None:
        self._set_alpha(value)

    def _set_alpha(self, value: Union[Tensor, float]) -> None:
        # Used by the alpha_prior
        if not isinstance(value, Tensor):
            value = torch.as_tensor(value).to(self.raw_alpha)
        self.initialize(raw_alpha=self.raw_alpha_constraint.inverse_transform(value))

    @property
    def beta(self) -> Tensor:
        return self.raw_beta_constraint.transform(self.raw_beta)

    @beta.setter
    def beta(self, value: Tensor) -> None:
        self._set_beta(value)

    def _set_beta(self, value: Union[Tensor, float]) -> None:
        # Used by the beta_prior
        if not isinstance(value, Tensor):
            value = torch.as_tensor(value).to(self.raw_beta)
        self.initialize(raw_beta=self.raw_beta_constraint.inverse_transform(value))

    def forward(self, x1: Tensor, x2: Tensor, diag: Optional[bool] = False, **params) -> Tensor:

        x1_, x2_ = x1.clone(), x2.clone()
        # Jitter datapoints that are exactly 0
        x1_[x1_ == 0], x2_[x2_ == 0] = x1_[x1_ == 0] + self.eps, x2_[x2_ == 0] + self.eps
        r1, r2 = x1_.norm(dim=-1, keepdim=True), x2_.norm(dim=-1, keepdim=True)

        if torch.any(r1 > 1.0) or torch.any(r2 > 1.0):
            raise RuntimeError("Cylindrical kernel not defined for data points with radius > 1. Scale your data!")

        a1, a2 = x1.div(r1), x2.div(r2)
        if not diag:
            gram_mat = a1.matmul(a2.transpose(-2, -1))
            for p in range(self.num_angular_weights):
                if p == 0:
                    angular_kernel = self.angular_weights[..., 0, None, None]
                else:
                    angular_kernel = angular_kernel + self.angular_weights[..., p, None, None].mul(gram_mat.pow(p))
        else:
            gram_mat = a1.mul(a2).sum(-1)
            for p in range(self.num_angular_weights):
                if p == 0:
                    angular_kernel = self.angular_weights[..., 0, None]
                else:
                    angular_kernel = angular_kernel + self.angular_weights[..., p, None].mul(gram_mat.pow(p))

        with settings.lazily_evaluate_kernels(False):
            radial_kernel = self.radial_base_kernel(self.kuma(r1), self.kuma(r2), diag=diag, **params)
        return radial_kernel.mul(angular_kernel)

    def kuma(self, x: Tensor) -> Tensor:
        alpha = self.alpha.view(*self.batch_shape, 1, 1)
        beta = self.beta.view(*self.batch_shape, 1, 1)

        res = 1 - (1 - x.pow(alpha) + self.eps).pow(beta)
        return res

    def num_outputs_per_input(self, x1: Tensor, x2: Tensor) -> int:
        return self.radial_base_kernel.num_outputs_per_input(x1, x2)



#!/usr/bin/env python3

import unittest
from functools import reduce
from itertools import combinations
from operator import mul

import torch
from linear_operator import to_dense

import gpytorch
from gpytorch.test.base_test_case import BaseTestCase


def prod(iterable):
    return reduce(mul, iterable, 1)


class TestSumInteractionTerms(BaseTestCase, unittest.TestCase):
    def test_sum_interaction_terms(self):
        batch_shape = torch.Size([2, 1])
        D = 5
        M = 4
        N = 20

        base_kernel = gpytorch.kernels.RBFKernel(batch_shape=torch.Size([D]))
        x = torch.randn(*batch_shape, D, N, 1)
        with torch.no_grad(), gpytorch.settings.lazily_evaluate_kernels(False):
            covars = base_kernel(x)

        actual = torch.zeros(*batch_shape, N, N)
        for degree in range(1, M + 1):
            for interaction_term_indices in combinations(range(D), degree):
                actual = actual + prod([to_dense(covars[..., i, :, :]) for i in interaction_term_indices])

        res = gpytorch.utils.sum_interaction_terms(covars, max_degree=M)
        self.assertAllClose(res, actual)



#!/usr/bin/env python3

from ..approximate_gp import ApproximateGP


class BayesianGPLVM(ApproximateGP):
    """
    The Gaussian Process Latent Variable Model (GPLVM) class for unsupervised learning.
    The class supports

    1. Point estimates for latent X when prior_x = None
    2. MAP Inference for X when prior_x is not None and inference == 'map'
    3. Gaussian variational distribution q(X) when prior_x is not None and inference == 'variational'

    .. seealso::
        The `GPLVM tutorial
        <examples/04_Variational_and_Approximate_GPs/Gaussian_Process_Latent_Variable_Models_with_Stochastic_Variational_Inference.ipynb>`_
        for use instructions.

    :param X: An instance of a sub-class of the LatentVariable class. One of,
        :class:`~gpytorch.models.gplvm.PointLatentVariable`, :class:`~gpytorch.models.gplvm.MAPLatentVariable`, or
        :class:`~gpytorch.models.gplvm.VariationalLatentVariable`, to facilitate inference with 1, 2, or 3 respectively.
    :type X: ~gpytorch.models.LatentVariable
    :param ~gpytorch.variational._VariationalStrategy variational_strategy: The strategy that determines
        how the model marginalizes over the variational distribution (over inducing points)
        to produce the approximate posterior distribution (over data)
    """

    def __init__(self, X, variational_strategy):
        super().__init__(variational_strategy)

        # Assigning Latent Variable
        self.X = X

    def forward(self):
        raise NotImplementedError

    def sample_latent_variable(self):
        sample = self.X()
        return sample



#!/usr/bin/env python3

import unittest

import torch

import gpytorch


class TestGrid(unittest.TestCase):
    def test_scale_to_bounds(self):
        """ """
        x = torch.randn(100) * 50
        scale_module = gpytorch.utils.grid.ScaleToBounds(-1.0, 1.0)
        res = scale_module(x)
        self.assertGreater(res.min().item(), -1.0)
        self.assertLess(res.max().item(), 1.0)

    def test_choose_grid_size(self):
        """ """
        x = torch.randn(100)
        grid_size = gpytorch.utils.grid.choose_grid_size(x, ratio=2.0)
        self.assertEqual(grid_size, 200)

        x = torch.randn(100, 1)
        grid_size = gpytorch.utils.grid.choose_grid_size(x, ratio=2.0)
        self.assertEqual(grid_size, 200)

        x = torch.randn(10000, 2)
        grid_size = gpytorch.utils.grid.choose_grid_size(x, ratio=2.0)
        self.assertEqual(grid_size, 200)

        x = torch.randn(16, 10000, 4)
        grid_size = gpytorch.utils.grid.choose_grid_size(x, ratio=2.0)
        self.assertEqual(grid_size, 20)



from .metrics import (
    mean_absolute_error,
    mean_squared_error,
    mean_standardized_log_loss,
    negative_log_predictive_density,
    quantile_coverage_error,
    standardized_mean_squared_error,
)

__all__ = [
    "mean_absolute_error",
    "mean_squared_error",
    "standardized_mean_squared_error",
    "mean_standardized_log_loss",
    "negative_log_predictive_density",
    "quantile_coverage_error",
]



#!/usr/bin/env python3

import unittest

import torch

from gpytorch import settings
from gpytorch.test.base_test_case import BaseTestCase


class TestSettings(BaseTestCase, unittest.TestCase):
    def test_feature_flag(self):
        self.assertTrue(settings.fast_pred_var.is_default())
        self.assertFalse(settings.fast_pred_var.on())
        with settings.fast_pred_var():
            self.assertFalse(settings.fast_pred_var.is_default())
            self.assertTrue(settings.fast_pred_var.on())
        with settings.fast_pred_var(False):
            self.assertFalse(settings.fast_pred_var.is_default())
            self.assertFalse(settings.fast_pred_var.on())

    def test_dtype_value_context(self):
        # test custom settings
        x = torch.zeros(1, dtype=torch.float)
        with settings.min_fixed_noise(float_value=0.1, double_value=0.2, half_value=0.3):
            self.assertEqual(settings.min_fixed_noise.value(x), 0.1)
            self.assertEqual(settings.min_fixed_noise.value(x.double()), 0.2)
            self.assertEqual(settings.min_fixed_noise.value(x.half()), 0.3)
        # test defaults are restored
        self.assertEqual(
            settings.min_fixed_noise.value(x),
            settings.min_fixed_noise._global_float_value,
        )
        self.assertEqual(
            settings.min_fixed_noise.value(x.double()),
            settings.min_fixed_noise._global_double_value,
        )
        self.assertEqual(
            settings.min_fixed_noise.value(x.half()),
            settings.min_fixed_noise._global_half_value,
        )
        # test setting one dtype
        with settings.min_fixed_noise(double_value=0.2):
            self.assertEqual(settings.min_fixed_noise.value(x.double()), 0.2)
            self.assertEqual(
                settings.min_fixed_noise.value(x),
                settings.min_fixed_noise._global_float_value,
            )
            self.assertEqual(
                settings.min_fixed_noise.value(x.half()),
                settings.min_fixed_noise._global_half_value,
            )



#!/usr/bin/env python3

import unittest

import torch

from gpytorch.utils.nearest_neighbors import NNUtil


class TestNNUtil(unittest.TestCase):
    def _setup(self, k, dim, batch_shape, train_n, test_n, preferred_nnlib, seq_train_n=None):
        # initialize
        nn_util = NNUtil(k=k, dim=dim, batch_shape=batch_shape, preferred_nnlib=preferred_nnlib)

        self._run(nn_util, dim, batch_shape, train_n, test_n, seq_train_n)
        return nn_util

    def _run(self, nn_util, dim, batch_shape, train_n, test_n, seq_train_n=None):
        train_x = torch.randn(*batch_shape, train_n, dim)
        test_x = torch.randn(*batch_shape, test_n, dim)
        k = nn_util.k

        # set training data
        nn_util.set_nn_idx(train_x)

        # test nn search
        test_nn_indices = nn_util.find_nn_idx(test_x)
        self.assertEqual(test_nn_indices.shape, (*batch_shape, test_n, k))

        # build sequential nn structure
        if seq_train_n is None:
            seq_train_x = train_x
            seq_train_n = train_n
        else:
            seq_train_x = torch.randn(*batch_shape, seq_train_n, dim)
        sequential_nn_idx = nn_util.build_sequential_nn_idx(seq_train_x)
        self.assertEqual(sequential_nn_idx.shape, (*batch_shape, seq_train_n - k, k))

    def test_setup(self):
        train_n = 10
        test_n = 5
        dim = 2
        k = 3

        new_train_n = 20
        new_test_n = 15

        for preferred_nnlib in ["sklearn", "faiss"]:
            nn_util = self._setup(
                k=k,
                dim=dim,
                batch_shape=torch.Size([]),
                train_n=train_n,
                test_n=test_n,
                preferred_nnlib=preferred_nnlib,
            )
            self._run(nn_util, dim=dim, batch_shape=nn_util.batch_shape, train_n=new_train_n, test_n=new_test_n)

            self._setup(
                k=k,
                dim=dim,
                batch_shape=torch.Size([2]),
                train_n=train_n,
                test_n=test_n,
                preferred_nnlib=preferred_nnlib,
            )
            self._run(nn_util, dim=dim, batch_shape=nn_util.batch_shape, train_n=new_train_n, test_n=new_test_n)

            self._setup(
                k=k,
                dim=dim,
                batch_shape=torch.Size([2, 5, 10]),
                train_n=train_n,
                test_n=test_n,
                seq_train_n=25,
                preferred_nnlib=preferred_nnlib,
            )
            self._run(
                nn_util,
                dim=dim,
                batch_shape=nn_util.batch_shape,
                train_n=new_train_n,
                test_n=new_test_n,
                seq_train_n=30,
            )

    def test_1Dtoycase(self):
        for preferred_nnlib in ["sklearn", "faiss"]:
            train_x = torch.arange(5, dtype=torch.float32).unsqueeze(-1)

            nn_util = NNUtil(k=3, dim=train_x.size(-1), preferred_nnlib=preferred_nnlib)

            # set train data
            nn_util.set_nn_idx(train_x)
            nn_idx = nn_util.find_nn_idx(train_x)
            nn_idx = torch.sort(nn_idx, dim=-1)[0]
            expected_idx = torch.tensor([[0, 1, 2], [0, 1, 2], [1, 2, 3], [2, 3, 4], [2, 3, 4]], dtype=torch.int64)
            assert torch.equal(
                nn_idx, expected_idx
            ), f"Preferred_nnlib = {preferred_nnlib}, nn_idx={nn_idx}, expected_idx={expected_idx}"

            # single point test
            test_x = torch.tensor([20], dtype=torch.float32)  # shape would be automatically expanded
            nn_idx = nn_util.find_nn_idx(test_x)
            nn_idx = torch.sort(nn_idx, dim=-1)[0]
            expected_idx = torch.tensor([[2, 3, 4]], dtype=torch.int64)
            assert torch.equal(nn_idx, expected_idx)

            # multiple point test
            test_x = torch.tensor([1.6, 3], dtype=torch.float32)  # shape would be automatically expanded
            nn_idx = nn_util.find_nn_idx(test_x)
            nn_idx = torch.sort(nn_idx, dim=-1)[0]
            expected_idx = torch.tensor([[1, 2, 3], [2, 3, 4]], dtype=torch.int64)
            assert torch.equal(nn_idx, expected_idx)

            # with less k
            nn_idx = nn_util.find_nn_idx(test_x, k=1)
            expected_idx = torch.tensor([[2], [3]], dtype=torch.int64)
            assert torch.equal(nn_idx, expected_idx)

            # build sequential nn idx
            seq_nn_idx = nn_util.build_sequential_nn_idx(train_x)
            seq_nn_idx = torch.sort(seq_nn_idx, dim=-1)[0]
            expected_idx = torch.tensor([[0, 1, 2], [1, 2, 3]], dtype=torch.int64)
            assert torch.equal(expected_idx, seq_nn_idx)

            # reset
            # set new train data
            train_x = torch.tensor([5.0, 3, 1, 4]).unsqueeze(-1)
            nn_util.set_nn_idx(train_x)

            test_x = torch.tensor([2.0, 10])
            nn_idx = nn_util.find_nn_idx(test_x)
            nn_idx = torch.sort(nn_idx, dim=-1)[0]
            expected_idx = torch.tensor([[1, 2, 3], [0, 1, 3]], dtype=torch.int64)
            assert torch.equal(nn_idx, expected_idx)

            seq_train_x = torch.tensor([10.0, 2, 5, 6, 9]).unsqueeze(-1)
            seq_nn_idx = nn_util.build_sequential_nn_idx(seq_train_x)
            seq_nn_idx = torch.sort(seq_nn_idx, dim=-1)[0]
            expected_idx = torch.tensor([[0, 1, 2], [0, 2, 3]], dtype=torch.int64)
            assert torch.equal(seq_nn_idx, expected_idx)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3

import unittest

from gpytorch.kernels import IndexKernel
from gpytorch.priors import NormalPrior


class TestIndexKernel(unittest.TestCase):
    def create_kernel_with_prior(self, prior):
        return IndexKernel(num_tasks=1, prior=prior)

    def test_prior_type(self):
        """
        Raising TypeError if prior type is other than gpytorch.priors.Prior
        """
        kernel_fn = lambda prior: self.create_kernel_with_prior(prior)
        kernel_fn(None)
        kernel_fn(NormalPrior(0, 1))
        self.assertRaises(TypeError, kernel_fn, 1)



import math
from typing import Optional

import torch
from torch import Tensor

from .kernel import Kernel


def _fmax(r: Tensor, j: int, q: int) -> Tensor:
    return torch.max(torch.tensor(0.0, dtype=r.dtype, device=r.device), 1 - r).pow(j + q)


def _get_cov(r: Tensor, j: int, q: int) -> Tensor:
    if q == 0:
        return 1
    if q == 1:
        return (j + 1) * r + 1
    if q == 2:
        return 1 + (j + 2) * r + ((j + 4 * j + 3) / 3.0) * (r**2)
    if q == 3:
        return (
            1
            + (j + 3) * r
            + ((6 * j**2 + 36 * j + 45) / 15.0) * r.square()
            + ((j**3 + 9 * j**2 + 23 * j + 15) / 15.0) * (r**3)
        )


class PiecewisePolynomialKernel(Kernel):
    r"""
    Computes a covariance matrix based on the Piecewise Polynomial kernel
    between inputs :math:`\mathbf{x_1}` and :math:`\mathbf{x_2}`:

    .. math::

        \begin{align}
            r &= \left\Vert x1 - x2 \right\Vert \\
            j &= \lfloor \frac{D}{2} \rfloor + q +1 \\
            K_{\text{ppD, 0}}(\mathbf{x_1}, \mathbf{x_2}) &= (1-r)^j_+ , \\
            K_{\text{ppD, 1}}(\mathbf{x_1}, \mathbf{x_2}) &= (1-r)^{j+1}_+ ((j + 1)r + 1), \\
            K_{\text{ppD, 2}}(\mathbf{x_1}, \mathbf{x_2}) &= (1-r)^{j+2}_+ ((1 + (j+2)r +
                \frac{j^2 + 4j + 3}{3}r^2), \\
            K_{\text{ppD, 3}}(\mathbf{x_1}, \mathbf{x_2}) &= (1-r)^{j+3}_+
                (1 + (j+3)r + \frac{6j^2 + 36j + 45}{15}r^2 +
                \frac{j^3 + 9j^2 + 23j +15}{15}r^3) \\
        \end{align}

    where :math:`K_{\text{ppD, q}}` is positive semidefinite in :math:`\mathbb{R}^{D}` and
    :math:`q` is the smoothness coefficient. See `Rasmussen and Williams (2006)`_ Equation 4.21.

    .. note:: This kernel does not have an `outputscale` parameter. To add a scaling parameter,
        decorate this kernel with a :class:`gpytorch.kernels.ScaleKernel`.

    :param int q: (default= 2) The smoothness parameter.
    :type q: int (0, 1, 2 or 3)
    :param ard_num_dims: (Default: `None`) Set this if you want a separate lengthscale for each
        input dimension. It should be `d` if x1 is a `... x n x d` matrix.
    :type ard_num_dims: int, optional
    :param batch_shape: (Default: `None`) Set this if you want a separate lengthscale for each
         batch of input data. It should be `torch.Size([b1, b2])` for a `b1 x b2 x n x m` kernel output.
    :type batch_shape: torch.Size, optional
    :param active_dims: (Default: `None`) Set this if you want to
        compute the covariance of only a few input dimensions. The ints
        corresponds to the indices of the dimensions.
    :type active_dims: Tuple(int)
    :param lengthscale_prior: (Default: `None`)
        Set this if you want to apply a prior to the lengthscale parameter.
    :type lengthscale_prior: ~gpytorch.priors.Prior, optional
    :param lengthscale_constraint: (Default: `Positive`) Set this if you want
        to apply a constraint to the lengthscale parameter.
    :type lengthscale_constraint: ~gpytorch.constraints.Positive, optional
    :param eps: (Default: 1e-6) The minimum value that the lengthscale can take (prevents divide by zero errors).
    :type eps: float, optional

    .. _Rasmussen and Williams (2006):
        http://www.gaussianprocess.org/gpml/

    Example:
        >>> x = torch.randn(10, 5)
        >>> # Non-batch option
        >>> covar_module = gpytorch.kernels.ScaleKernel(
                                gpytorch.kernels.PiecewisePolynomialKernel(q = 2))
        >>> # Non-batch: ARD (different lengthscale for each input dimension)
        >>> covar_module = gpytorch.kernels.ScaleKernel(
                            gpytorch.kernels.PiecewisePolynomialKernel(q = 2, ard_num_dims=5)
                            )
        >>> covar = covar_module(x)  # Output: LinearOperator of size (10 x 10)
        >>> batch_x = torch.randn(2, 10, 5)
        >>> # Batch: different lengthscale for each batch
        >>> covar_module = gpytorch.kernels.ScaleKernel(
            gpytorch.kernels.PiecewisePolynomialKernel(q = 2, batch_shape=torch.Size([2]))
            )
        >>> covar = covar_module(batch_x)  # Output: LinearOperator of size (2 x 10 x 10)
    """
    has_lengthscale = True

    def __init__(self, q: Optional[int] = 2, **kwargs):
        super(PiecewisePolynomialKernel, self).__init__(**kwargs)
        if q not in {0, 1, 2, 3}:
            raise ValueError("q expected to be 0, 1, 2 or 3")
        self.q = q

    def forward(self, x1: Tensor, x2: Tensor, last_dim_is_batch: bool = False, diag: bool = False, **params) -> Tensor:
        x1_ = x1.div(self.lengthscale)
        x2_ = x2.div(self.lengthscale)
        if last_dim_is_batch is True:
            D = x1.shape[1]
        else:
            D = x1.shape[-1]
        j = math.floor(D / 2.0) + self.q + 1
        if last_dim_is_batch and diag:
            r = self.covar_dist(x1_, x2_, last_dim_is_batch=True, diag=True)
        elif diag:
            r = self.covar_dist(x1_, x2_, diag=True)
        elif last_dim_is_batch:
            r = self.covar_dist(x1_, x2_, last_dim_is_batch=True)
        else:
            r = self.covar_dist(x1_, x2_)
        cov_matrix = _fmax(r, j, self.q) * _get_cov(r, j, self.q)
        return cov_matrix



#!/usr/bin/env python3

from typing import Optional, Union

import torch
from linear_operator.operators import (
    LinearOperator,
    LowRankRootLinearOperator,
    MatmulLinearOperator,
    RootLinearOperator,
)
from torch import Tensor

from ..constraints import Interval, Positive
from ..models import exact_prediction_strategies
from ..priors import Prior
from .kernel import Kernel


class LinearKernel(Kernel):
    r"""
    Computes a covariance matrix based on the Linear kernel
    between inputs :math:`\mathbf{x_1}` and :math:`\mathbf{x_2}`:

    .. math::
        \begin{equation*}
            k_\text{Linear}(\mathbf{x_1}, \mathbf{x_2}) = v\mathbf{x_1}^\top
            \mathbf{x_2}.
        \end{equation*}

    where

    * :math:`v` is a variance parameter.


    .. note::

        To implement this efficiently, we use a
        :obj:`~linear_operator.operators.RootLinearOperator` during training
        and a :class:`~linear_operator.operators.MatmulLinearOperator` during
        test. These lazy tensors represent matrices of the form :math:`\mathbf
        K = \mathbf X \mathbf X^{\prime \top}`. This makes inference efficient
        because a matrix-vector product :math:`\mathbf K \mathbf v` can be
        computed as :math:`\mathbf K \mathbf v = \mathbf X( \mathbf X^{\prime
        \top} \mathbf v)`, where the base multiply :math:`\mathbf X \mathbf v`
        takes only :math:`\mathcal O(ND)` time and space.

    :param ard_num_dims: Set this if you want a separate variance priors for each weight. (Default: `None`)
    :param variance_prior: Prior over the variance parameter. (Default `None`.)
    :param variance_constraint: Constraint to place on variance parameter. (Default: `Positive`.)
    :param active_dims: List of data dimensions to operate on.
    """

    def __init__(
        self,
        ard_num_dims: Optional[int] = None,
        variance_prior: Optional[Prior] = None,
        variance_constraint: Optional[Interval] = None,
        **kwargs,
    ):
        super().__init__(ard_num_dims=ard_num_dims, **kwargs)
        if variance_constraint is None:
            variance_constraint = Positive()
        self.register_parameter(
            name="raw_variance",
            parameter=torch.nn.Parameter(
                torch.zeros(*self.batch_shape, 1, 1 if ard_num_dims is None else ard_num_dims)
            ),
        )
        if variance_prior is not None:
            if not isinstance(variance_prior, Prior):
                raise TypeError("Expected gpytorch.priors.Prior but got " + type(variance_prior).__name__)
            self.register_prior("variance_prior", variance_prior, lambda m: m.variance, lambda m, v: m._set_variance(v))

        self.register_constraint("raw_variance", variance_constraint)

    @property
    def variance(self) -> Tensor:
        return self.raw_variance_constraint.transform(self.raw_variance)

    @variance.setter
    def variance(self, value: Union[float, Tensor]) -> None:
        self._set_variance(value)

    def _set_variance(self, value: Union[float, Tensor]) -> None:
        if not torch.is_tensor(value):
            value = torch.as_tensor(value).to(self.raw_variance)
        self.initialize(raw_variance=self.raw_variance_constraint.inverse_transform(value))

    def forward(
        self, x1: Tensor, x2: Tensor, diag: bool = False, last_dim_is_batch: Optional[bool] = False, **params
    ) -> Union[Tensor, LinearOperator]:
        x1_ = x1 * self.variance.sqrt()
        if last_dim_is_batch:
            x1_ = x1_.transpose(-1, -2).unsqueeze(-1)

        if x1.size() == x2.size() and torch.equal(x1, x2):
            # Use RootLinearOperator when x1 == x2 for efficiency when composing
            # with other kernels
            n, d = x1.shape[-2:]
            prod = RootLinearOperator(x1_) if d > n else LowRankRootLinearOperator(x1_)

        else:
            x2_ = x2 * self.variance.sqrt()
            if last_dim_is_batch:
                x2_ = x2_.transpose(-1, -2).unsqueeze(-1)

            prod = MatmulLinearOperator(x1_, x2_.transpose(-2, -1))

        if diag:
            return prod.diagonal(dim1=-1, dim2=-2)
        else:
            return prod

    def prediction_strategy(self, train_inputs, train_prior_dist, train_labels, likelihood):
        # Allow for fast sampling
        return exact_prediction_strategies.LinearPredictionStrategy(
            train_inputs, train_prior_dist, train_labels, likelihood
        )



#!/usr/bin/env python3

from __future__ import annotations

import warnings as _warnings
from typing import Any

import linear_operator

from . import deprecation, errors, generic, grid, interpolation, quadrature, transforms, warnings
from .memoize import cached
from .nearest_neighbors import NNUtil
from .sum_interaction_terms import sum_interaction_terms

__all__ = [
    "cached",
    "deprecation",
    "errors",
    "generic",
    "grid",
    "interpolation",
    "quadrature",
    "sum_interaction_terms",
    "transforms",
    "warnings",
    "NNUtil",
]


def __getattr__(name: str) -> Any:
    if hasattr(linear_operator.utils, name):
        _warnings.warn(
            f"gpytorch.utils.{name} is deprecated. Use linear_operator.utils.{name} instead.",
            DeprecationWarning,
        )
        return getattr(linear_operator.utils, name)
    raise AttributeError(f"module gpytorch.utils has no attribute {name}")



#!/usr/bin/env python3

import unittest

import torch
from linear_operator.operators import ToeplitzLinearOperator
from torch.distributions import Distribution

import gpytorch
from gpytorch.distributions import MultivariateNormal

from gpytorch.likelihoods import GaussianLikelihood, HadamardGaussianLikelihood, Likelihood
from gpytorch.test.base_likelihood_test_case import BaseLikelihoodTestCase


class TestMultitaskGaussianLikelihood(BaseLikelihoodTestCase, unittest.TestCase):
    seed = 2

    def _create_conditional_input(self, batch_shape=torch.Size([])):
        return torch.randn(*batch_shape, 5)

    def _create_task_input(self, batch_shape=torch.Size([])):
        return torch.tensor([0, 1, 2, 3, 0], dtype=torch.long).expand(*batch_shape, 5).unsqueeze(-1)

    def _create_marginal_input(self, batch_shape=torch.Size([])):
        data_and_task_mat = ToeplitzLinearOperator(torch.tensor([1, 0.6, 0.4, 0.2, 0.1]))
        return MultivariateNormal(torch.randn(*batch_shape, 5), data_and_task_mat)

    def create_likelihood(self):
        return HadamardGaussianLikelihood(num_tasks=4)

    def _test_conditional(self, batch_shape):
        likelihood = self.create_likelihood()
        likelihood.max_plate_nesting += len(batch_shape)
        input = self._create_conditional_input(batch_shape)
        task_idcs = self._create_task_input(batch_shape)
        output = likelihood(input, [task_idcs])

        self.assertTrue(isinstance(output, Distribution))
        self.assertEqual(output.sample().shape, input.shape)

    def _test_log_marginal(self, batch_shape):
        likelihood = self.create_likelihood()
        likelihood.max_plate_nesting += len(batch_shape)
        input = self._create_marginal_input(batch_shape)
        task_idcs = self._create_task_input(batch_shape)
        target = self._create_targets(batch_shape)
        with gpytorch.settings.num_likelihood_samples(512):
            output = likelihood.log_marginal(target, input, [task_idcs])

        self.assertTrue(torch.is_tensor(output))
        self.assertEqual(output.shape, batch_shape + torch.Size([5]))

        with gpytorch.settings.num_likelihood_samples(512):
            # Since all tasks are initialized with the same noise, this is
            # equivalent to using a shared GaussianLikelihood
            default_log_prob = Likelihood.log_marginal(GaussianLikelihood(), target, input)
        self.assertAllClose(output, default_log_prob, rtol=0.25)

    def _test_log_prob(self, batch_shape):
        likelihood = self.create_likelihood()
        likelihood.max_plate_nesting += len(batch_shape)
        input = self._create_marginal_input(batch_shape)
        task_idcs = self._create_task_input(batch_shape)
        target = self._create_targets(batch_shape)
        with gpytorch.settings.num_likelihood_samples(512):
            output = likelihood.expected_log_prob(target, input, [task_idcs])

        self.assertTrue(torch.is_tensor(output))
        self.assertEqual(output.shape, batch_shape + torch.Size([5]))

        with gpytorch.settings.num_likelihood_samples(512):
            # Since all tasks are initialized with the same noise, this is
            # equivalent to using a shared GaussianLikelihood
            default_log_prob = Likelihood.expected_log_prob(GaussianLikelihood(), target, input)
        self.assertAllClose(output, default_log_prob, rtol=0.25)

    def _test_marginal(self, batch_shape):
        likelihood = self.create_likelihood()
        likelihood.max_plate_nesting += len(batch_shape)
        input = self._create_marginal_input(batch_shape)
        task_idcs = self._create_task_input(batch_shape)
        output = likelihood(input, [task_idcs])

        self.assertTrue(isinstance(output, Distribution))
        self.assertEqual(output.sample().shape[-len(input.sample().shape) :], input.sample().shape)

        # Compare against default implementation
        with gpytorch.settings.num_likelihood_samples(30000):
            default = Likelihood.marginal(likelihood, input, [task_idcs])
        # print(output.mean, default.mean)
        default_mean = default.mean
        actual_mean = output.mean
        if default_mean.dim() > actual_mean.dim():
            default_mean = default_mean.mean(0)
        self.assertAllClose(default_mean, actual_mean, rtol=0.25, atol=0.25)

    def test_marginal_variance(self):
        likelihood = HadamardGaussianLikelihood(num_tasks=4)
        likelihood.noise = torch.tensor([[0.1, 0.2, 0.3, 0.4]])

        input = self._create_marginal_input()
        task_idcs = self._create_task_input()
        variance = likelihood(input, [task_idcs]).variance
        self.assertAllClose(variance, torch.tensor([1.1, 1.2, 1.3, 1.4, 1.1]))

        # The shape and dtypes of the tasks must be (num_data, 1) and integer dtype
        with self.assertRaises(ValueError):
            likelihood(input)

        with self.assertRaises(ValueError):
            likelihood(input, [])

        with self.assertRaises(ValueError):
            likelihood(input, [task_idcs.squeeze(-1)])

        # test with task feature and full input
        likelihood = HadamardGaussianLikelihood(num_tasks=4, task_feature_index=1)
        likelihood.noise = torch.tensor([[0.1, 0.2, 0.3, 0.4]])
        X = torch.cat([torch.zeros_like(task_idcs), task_idcs], dim=-1)
        self.assertEqual(variance, likelihood(input, [X]).variance)



#!/usr/bin/env python3

import torch
from linear_operator.utils.cholesky import psd_safe_cholesky
from torch.distributions import constraints, LKJCholesky
from torch.nn import Module as TModule

from .. import settings
from .prior import Prior


class LKJCholeskyFactorPrior(Prior, LKJCholesky):
    r"""LKJ prior over n x n (positive definite) Cholesky-decomposed
    correlation matrices

    .. math:

        \begin{equation*}
            pdf(\Sigma) \sim |\Sigma| ^ (\eta  - 1)
        \end{equation*}

    where :math:`\eta > 0` is a shape parameter and n is the dimension of the
    correlation matrix.

    LKJCholeskyFactorPrior is different from LKJPrior in that it accepts the
    Cholesky factor of the correlation matrix to compute probabilities.

    This distribution assumes that the cholesky factor is lower-triangular.
    """

    def __init__(self, n, eta, validate_args=False, transform=None):
        TModule.__init__(self)
        LKJCholesky.__init__(self, dim=n, concentration=eta, validate_args=validate_args)
        self.n = self.dim
        self.eta = self.concentration
        self._transform = transform

    # TODO: change from sample to rsample if pytorch #69281 goes in providing rsample method


class LKJPrior(LKJCholeskyFactorPrior):
    r"""LKJ prior over n x n (positive definite) correlation matrices

    .. math:

        \begin{equation*}
            pdf(\Sigma) \sim |\Sigma| ^ (\eta  - 1)
        \end{equation*}

    where :math:`\eta > 0` is a shape parameter.

    Reference: Bayesian Data Analysis, 3rd ed., Gelman et al., p. 576
    """
    support = constraints.positive_definite

    def log_prob(self, X):
        if any(s != self.n for s in X.shape[-2:]):
            raise ValueError("Correlation matrix is not of size n={}".format(self.n))
        if not _is_valid_correlation_matrix(X):
            raise ValueError("Input is not a valid correlation matrix")
        X_cholesky = psd_safe_cholesky(X, upper=False)
        return super().log_prob(X_cholesky)

    def sample(self, sample_shape=torch.Size()):
        R = super().sample(sample_shape=sample_shape)
        return R.matmul(R.transpose(-1, -2))


class LKJCovariancePrior(LKJPrior):
    """LKJCovariancePrior combines an LKJ prior over the correlation matrix
    and a user-specified prior over marginal standard deviations to return a
    prior over the full covariance matrix.

    Usage: LKJCovariancePrior(n, eta, sd_prior), where
        n is a positive integer, the size of the covariance matrix,
        eta is a positive shape parameter for the LKJPrior over correlations, and
        sd_prior is a scalar Prior over nonnegative numbers, which is used for
        each of the n marginal standard deviations on the covariance matrix.
    """

    def __init__(self, n, eta, sd_prior, validate_args=False):
        if not isinstance(sd_prior, Prior):
            raise ValueError("sd_prior must be an instance of Prior")
        if not isinstance(n, int):
            raise ValueError("n must be an integer")
        # bug-fix event shapes if necessary
        if sd_prior.event_shape not in {torch.Size([1]), torch.Size([n]), torch.Size([])}:
            if sd_prior.event_shape == torch.Size([]):
                sd_prior._event_shape = torch.Size([1])
            raise ValueError("sd_prior must have event_shape 1 or n")
        correlation_prior = LKJPrior(n=n, eta=eta, validate_args=validate_args)
        if sd_prior.batch_shape != correlation_prior.batch_shape:
            raise ValueError("sd_prior must have same batch_shape as eta")
        TModule.__init__(self)
        self.correlation_prior = correlation_prior
        self.sd_prior = sd_prior

    def log_prob(self, X):
        marginal_var = torch.diagonal(X, dim1=-2, dim2=-1)
        if not torch.all(marginal_var >= 0):
            raise ValueError("Variance(s) cannot be negative")
        marginal_sd = marginal_var.sqrt()
        sd_diag_mat = _batch_form_diag(1 / marginal_sd)
        correlations = torch.matmul(torch.matmul(sd_diag_mat, X), sd_diag_mat)
        log_prob_corr = self.correlation_prior.log_prob(correlations)
        log_prob_sd = self.sd_prior.log_prob(marginal_sd)
        return log_prob_corr + log_prob_sd

    def sample(self, sample_shape=torch.Size()):
        base_correlation = self.correlation_prior.sample(sample_shape)
        marginal_sds = self.sd_prior.rsample(sample_shape)
        # expand sds to have the same shape as the base correlation matrix
        marginal_sds = marginal_sds.repeat(*[1] * len(sample_shape), self.correlation_prior.n)
        marginal_sds = torch.diag_embed(marginal_sds)
        return marginal_sds.matmul(base_correlation).matmul(marginal_sds)


def _batch_form_diag(tsr):
    """Form diagonal matrices in batch mode."""
    eye = torch.eye(tsr.shape[-1], dtype=tsr.dtype, device=tsr.device)
    M = tsr.unsqueeze(-1).expand(tsr.shape + tsr.shape[-1:])
    return eye * M


def _is_valid_correlation_matrix(Sigma, tol=1e-6):
    """Check if supplied matrix is a valid correlation matrix

    A matrix is a valid correlation matrix if it is positive semidefinite, and
    if all diagonal elements are equal to 1.

    Args:
        Sigma: A n x n correlation matrix, or a batch of b correlation matrices
            with shape b x n x n
        tol: The tolerance with which to check unit value of the diagonal elements

    Returns:
        True if Sigma is a valid correlation matrix, False otherwise (in batch
            mode, all matrices in the batch need to be valid correlation matrices)

    """
    if settings.verbose_linalg.on():
        settings.verbose_linalg.logger.debug(f"Running symeig on a matrix of size {Sigma.shape}.")

    evals = torch.linalg.eigvalsh(Sigma)
    if not torch.all(evals >= -tol):
        return False
    return all(torch.all(torch.abs(S.diagonal(dim1=-1, dim2=-2) - 1) < tol) for S in Sigma.view(-1, *Sigma.shape[-2:]))


def _is_valid_correlation_matrix_cholesky_factor(L, tol=1e-6):
    """Check if supplied matrix is a Cholesky factor of a valid correlation matrix

    A matrix is a Cholesky fator of a valid correlation matrix if it is lower
    triangular, has positive diagonal, and unit row-sum

    Args:
        L: A n x n lower-triangular matrix, or a batch of b lower-triangular
            matrices with shape b x n x n
        tol: The tolerance with which to check positivity of the diagonal and
            unit-sum of the rows

    Returns:
        True if L is a Cholesky factor of a valid correlation matrix, False
            otherwise (in batch mode, all matrices in the batch need to be
            Cholesky factors of valid correlation matrices)

    """
    unit_row_length = torch.all((torch.norm(L, dim=-1) - 1).abs() < tol)
    return unit_row_length and torch.all(constraints.lower_cholesky.check(L))



#!/usr/bin/env python3

import unittest
from math import pi

import torch

import gpytorch
from gpytorch.distributions import MultitaskMultivariateNormal
from gpytorch.likelihoods import MultitaskGaussianLikelihood
from gpytorch.means import ConstantMean, MultitaskMean
from gpytorch.test.base_test_case import BaseTestCase


# Simple training data: let's try to learn a sine function
train_x = torch.linspace(0, 1, 100)

# y1 function is sin(2*pi*x) with noise N(0, 0.04)
train_y1 = torch.sin(train_x * (2 * pi)) + torch.randn(train_x.size()) * 0.1
# y2 function is cos(2*pi*x) with noise N(0, 0.04)
train_y2 = torch.cos(train_x * (2 * pi)) + torch.randn(train_x.size()) * 0.1

# Create a train_y which interleaves the two
train_y = torch.stack([train_y1, train_y2], -1)


class MultitaskGPModel(gpytorch.models.ExactGP):
    def __init__(self, train_x, train_y, likelihood):
        super(MultitaskGPModel, self).__init__(train_x, train_y, likelihood)
        self.mean_module = MultitaskMean(ConstantMean(), num_tasks=2)
        self.covar_module = gpytorch.kernels.MultitaskKernel(
            gpytorch.kernels.InducingPointKernel(
                gpytorch.kernels.RBFKernel(), inducing_points=torch.randn(50, 1), likelihood=likelihood
            ),
            num_tasks=2,
            rank=2,
        )

    def forward(self, x):
        mean_x = self.mean_module(x)
        covar_x = self.covar_module(x)
        return MultitaskMultivariateNormal(mean_x, covar_x)


class TestSimpleGPRegression(BaseTestCase, unittest.TestCase):
    seed = 0

    def test_multitask_gp_mean_abs_error(self):
        likelihood = MultitaskGaussianLikelihood(num_tasks=2)
        model = MultitaskGPModel(train_x, train_y, likelihood)
        # Find optimal model hyperparameters
        model.train()
        likelihood.train()

        # Use the adam optimizer
        optimizer = torch.optim.Adam(model.parameters(), lr=0.1)  # Includes GaussianLikelihood parameters

        # "Loss" for GPs - the marginal log likelihood
        mll = gpytorch.mlls.ExactMarginalLogLikelihood(likelihood, model)

        n_iter = 50
        for _ in range(n_iter):
            # Zero prev backpropped gradients
            optimizer.zero_grad()
            # Make predictions from training data
            # Again, note feeding duplicated x_data and indices indicating which task
            output = model(train_x)
            # TODO: Fix this view call!!
            loss = -mll(output, train_y)
            loss.backward()
            optimizer.step()

        # Test the model
        model.eval()
        likelihood.eval()
        test_x = torch.linspace(0, 1, 51)
        test_y1 = torch.sin(test_x * (2 * pi))
        test_y2 = torch.cos(test_x * (2 * pi))
        test_preds = likelihood(model(test_x)).mean
        mean_abs_error_task_1 = torch.mean(torch.abs(test_y1 - test_preds[:, 0]))
        mean_abs_error_task_2 = torch.mean(torch.abs(test_y2 - test_preds[:, 1]))

        self.assertLess(mean_abs_error_task_1.squeeze().item(), 0.05)
        self.assertLess(mean_abs_error_task_2.squeeze().item(), 0.05)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3

from typing import Optional, Tuple, Union

import linear_operator
import torch
from linear_operator import LinearOperator
from torch import Tensor

from . import (
    beta_features,
    constraints,
    distributions,
    kernels,
    lazy,
    likelihoods,
    means,
    metrics,
    mlls,
    models,
    optim,
    priors,
    settings,
    utils,
    variational,
)
from .functions import inv_matmul, log_normal_cdf, logdet, matmul  # Deprecated
from .mlls import ExactMarginalLogLikelihood
from .module import Module

Anysor = Union[LinearOperator, Tensor]


def add_diagonal(input: Anysor, diag: Tensor) -> LinearOperator:
    r"""
    Adds an element to the diagonal of the matrix :math:`\mathbf A`.

    :param input: The matrix (or batch of matrices) :math:`\mathbf A` (... x N x N).
    :param diag: Diagonal to add
    :return: :math:`\mathbf A + \text{diag}(\mathbf d)`, where :math:`\mathbf A` is the linear operator
        and :math:`\mathbf d` is the diagonal component
    """
    return linear_operator.add_diagonal(input=input, diag=diag)


def add_jitter(input: Anysor, jitter_val: float = 1e-3) -> Anysor:
    r"""
    Adds jitter (i.e., a small diagonal component) to the matrix this
    LinearOperator represents.
    This is equivalent to calling :meth:`~linear_operator.operators.LinearOperator.add_diagonal`
    with a scalar tensor.

    :param input: The matrix (or batch of matrices) :math:`\mathbf A` (... x N x N).
    :param jitter_val: The diagonal component to add
    :return: :math:`\mathbf A + \alpha (\mathbf I)`, where :math:`\mathbf A` is the linear operator
        and :math:`\alpha` is :attr:`jitter_val`.
    """
    return linear_operator.add_jitter(input=input, jitter_val=jitter_val)


def diagonalization(input: Anysor, method: Optional[str] = None) -> Tuple[Tensor, Tensor]:
    r"""
    Returns a (usually partial) diagonalization of a symmetric positive definite matrix (or batch of matrices).
    :math:`\mathbf A`.
    Options are either "lanczos" or "symeig". "lanczos" runs Lanczos while
    "symeig" runs LinearOperator.symeig.

    :param input: The matrix (or batch of matrices) :math:`\mathbf A` (... x N x N).
    :param method: Specify the method to use ("lanczos" or "symeig"). The method will be determined
        based on size if not specified.
    :return: eigenvalues and eigenvectors representing the diagonalization.
    """
    return linear_operator.diagonalization(input=input, method=method)


def dsmm(
    sparse_mat: Union[torch.sparse.HalfTensor, torch.sparse.FloatTensor, torch.sparse.DoubleTensor],
    dense_mat: Tensor,
) -> Tensor:
    r"""
    Performs the (batch) matrix multiplication :math:`\mathbf{SD}`
    where :math:`\mathbf S` is a sparse matrix and :math:`\mathbf D` is a dense matrix.

    :param sparse_mat: Sparse matrix :math:`\mathbf S` (... x M x N)
    :param dense_mat: Dense matrix :math:`\mathbf D` (... x N x O)
    :return: :math:`\mathbf S \mathbf D` (... x M x N)
    """
    return linear_operator.dsmm(sparse_mat=sparse_mat, dense_mat=dense_mat)


def inv_quad(input: Anysor, inv_quad_rhs: Tensor, reduce_inv_quad: bool = True) -> Tensor:
    r"""
    Computes an inverse quadratic form (w.r.t self) with several right hand sides, i.e:

    .. math::
       \text{tr}\left( \mathbf R^\top \mathbf A^{-1} \mathbf R \right),

    where :math:`\mathbf A` is a positive definite matrix (or batch of matrices) and :math:`\mathbf R`
    represents the right hand sides (:attr:`inv_quad_rhs`).

    If :attr:`reduce_inv_quad` is set to false (and :attr:`inv_quad_rhs` is supplied),
    the function instead computes

    .. math::
       \text{diag}\left( \mathbf R^\top \mathbf A^{-1} \mathbf R \right).

    :param input: :math:`\mathbf A` - the positive definite matrix (... X N X N)
    :param inv_quad_rhs: :math:`\mathbf R` - the right hand sides of the inverse quadratic term (... x N x M)
    :param reduce_inv_quad: Whether to compute
        :math:`\text{tr}\left( \mathbf R^\top \mathbf A^{-1} \mathbf R \right)`
        or :math:`\text{diag}\left( \mathbf R^\top \mathbf A^{-1} \mathbf R \right)`.
    :returns: The inverse quadratic term.
        If `reduce_inv_quad=True`, the inverse quadratic term is of shape (...). Otherwise, it is (... x M).
    """
    return linear_operator.inv_quad(input=input, inv_quad_rhs=inv_quad_rhs, reduce_inv_quad=reduce_inv_quad)


def inv_quad_logdet(
    input: Anysor,
    inv_quad_rhs: Optional[Tensor] = None,
    logdet: bool = False,
    reduce_inv_quad: bool = True,
) -> Tuple[Tensor, Tensor]:
    r"""
    Calls both :func:`inv_quad_logdet` and :func:`logdet` on a positive definite matrix (or batch) :math:`\mathbf A`.
    However, calling this method is far more efficient and stable than calling each method independently.

    :param input: :math:`\mathbf A` - the positive definite matrix (... X N X N)
    :param inv_quad_rhs: :math:`\mathbf R` - the right hand sides of the inverse quadratic term (... x N x M)
    :param logdet: Whether or not to compute the
        logdet term :math:`\log \vert \mathbf A \vert`.
    :param reduce_inv_quad: Whether to compute
        :math:`\text{tr}\left( \mathbf R^\top \mathbf A^{-1} \mathbf R \right)`
        or :math:`\text{diag}\left( \mathbf R^\top \mathbf A^{-1} \mathbf R \right)`.
    :returns: The inverse quadratic term (or None), and the logdet term (or None).
        If `reduce_inv_quad=True`, the inverse quadratic term is of shape (...). Otherwise, it is (... x M).
    """
    return linear_operator.inv_quad_logdet(
        input=input,
        inv_quad_rhs=inv_quad_rhs,
        logdet=logdet,
        reduce_inv_quad=reduce_inv_quad,
    )


def pivoted_cholesky(
    input: Anysor,
    rank: int,
    error_tol: Optional[float] = None,
    return_pivots: bool = False,
) -> Union[Tensor, Tuple[Tensor, Tensor]]:
    r"""
    Performs a partial pivoted Cholesky factorization of a positive definite matrix (or batch of matrices).
    :math:`\mathbf L \mathbf L^\top = \mathbf A`.
    The partial pivoted Cholesky factor :math:`\mathbf L \in \mathbb R^{N \times \text{rank}}`
    forms a low rank approximation to the LinearOperator.

    The pivots are selected greedily, correspoading to the maximum diagonal element in the
    residual after each Cholesky iteration. See `Harbrecht et al., 2012`_.

    :param input: The matrix (or batch of matrices) :math:`\mathbf A` (... x N x N).
    :param rank: The size of the partial pivoted Cholesky factor.
    :param error_tol: Defines an optional stopping criterion.
        If the residual of the factorization is less than :attr:`error_tol`, then the
        factorization will exit early. This will result in a :math:`\leq \text{ rank}` factor.
    :param return_pivots: Whether or not to return the pivots alongside
        the partial pivoted Cholesky factor.
    :return: The `... x N x rank` factor (and optionally the `... x N` pivots if :attr:`return_pivots` is True).

    .. _Harbrecht et al., 2012:
        https://www.sciencedirect.com/science/article/pii/S0168927411001814
    """
    return linear_operator.pivoted_cholesky(input=input, rank=rank, return_pivots=return_pivots)


def root_decomposition(input: Anysor, method: Optional[str] = None) -> LinearOperator:
    r"""
    Returns a (usually low-rank) root decomposition linear operator of the
    positive definite matrix (or batch of matrices) :math:`\mathbf A`.
    This can be used for sampling from a Gaussian distribution, or for obtaining a
    low-rank version of a matrix.

    :param input: The matrix (or batch of matrices) :math:`\mathbf A` (... x N x N).
    :param method: Which method to use to perform the root decomposition. Choices are:
        "cholesky", "lanczos", "symeig", "pivoted_cholesky", or "svd".
    :return: A tensor :math:`\mathbf R` such that :math:`\mathbf R \mathbf R^\top \approx \mathbf A`.
    """
    return linear_operator.root_decomposition(input=input, method=method)


def root_inv_decomposition(
    input: Anysor,
    initial_vectors: Optional[Tensor] = None,
    test_vectors: Optional[Tensor] = None,
    method: Optional[str] = None,
) -> LinearOperator:
    r"""
    Returns a (usually low-rank) inverse root decomposition linear operator
    of the PSD LinearOperator :math:`\mathbf A`.
    This can be used for sampling from a Gaussian distribution, or for obtaining a
    low-rank version of a matrix.

    The root_inv_decomposition is performed using a partial Lanczos tridiagonalization.

    :param input: The matrix (or batch of matrices) :math:`\mathbf A` (... x N x N).
    :param initial_vectors: Vectors used to initialize the Lanczos decomposition.
        The best initialization vector (determined by :attr:`test_vectors`) will be chosen.
    :param test_vectors: Vectors used to test the accuracy of the decomposition.
    :param method: Root decomposition method to use (symeig, diagonalization, lanczos, or cholesky).
    :return: A tensor :math:`\mathbf R` such that :math:`\mathbf R \mathbf R^\top \approx \mathbf A^{-1}`.
    """
    return linear_operator.root_inv_decomposition(
        input=input,
        initial_vectors=initial_vectors,
        test_vectors=test_vectors,
        method=method,
    )


def solve(input: Anysor, rhs: Tensor, lhs: Optional[Tensor] = None) -> Tensor:
    r"""
    Given a positive definite matrix (or batch of matrices) :math:`\mathbf A`,
    computes a linear solve with right hand side :math:`\mathbf R`:

    .. math::
       \begin{equation}
           \mathbf A^{-1} \mathbf R,
       \end{equation}

    where :math:`\mathbf R` is :attr:`right_tensor` and :math:`\mathbf A` is the LinearOperator.

    .. note::
        Unlike :func:`torch.linalg.solve`, this function can take an optional :attr:`left_tensor` attribute.
        If this is supplied :func:`gpytorch.solve` computes

        .. math::
           \begin{equation}
               \mathbf L \mathbf A^{-1} \mathbf R,
           \end{equation}

        where :math:`\mathbf L` is :attr:`left_tensor`.
        Supplying this can reduce the number of solver calls required in the backward pass.

    :param input: The matrix (or batch of matrices) :math:`\mathbf A` (... x N x N).
    :param rhs: :math:`\mathbf R` - the right hand side
    :param lhs: :math:`\mathbf L` - the left hand side
    :return: :math:`\mathbf A^{-1} \mathbf R` or :math:`\mathbf L \mathbf A^{-1} \mathbf R`.
    """
    return linear_operator.solve(input=input, rhs=rhs, lhs=lhs)


def sqrt_inv_matmul(input: Anysor, rhs: Tensor, lhs: Optional[Tensor] = None) -> Tensor:
    r"""
    Given a positive definite matrix (or batch of matrices) :math:`\mathbf A`
    and a right hand size :math:`\mathbf R`,
    computes

    .. math::
       \begin{equation}
           \mathbf A^{-1/2} \mathbf R,
       \end{equation}

    If :attr:`lhs` is supplied, computes

    .. math::
       \begin{equation}
           \mathbf L \mathbf A^{-1/2} \mathbf R,
       \end{equation}

    where :math:`\mathbf L` is :attr:`lhs`.
    (Supplying :attr:`lhs` can reduce the number of solver calls required in the backward pass.)

    :param input: The matrix (or batch of matrices) :math:`\mathbf A` (... x N x N).
    :param rhs: :math:`\mathbf R` - the right hand side
    :param lhs: :math:`\mathbf L` - the left hand side
    :return: :math:`\mathbf A^{-1/2} \mathbf R` or :math:`\mathbf L \mathbf A^{-1/2} \mathbf R`.
    """
    return linear_operator.sqrt_inv_matmul(input=input, rhs=rhs, lhs=lhs)


# Read version number as written by setuptools_scm
try:
    from gpytorch.version import version as __version__
except Exception:  # pragma: no cover
    __version__ = "Unknown"  # pragma: no cover


__all__ = [
    # Submodules
    "constraints",
    "distributions",
    "kernels",
    "lazy",
    "likelihoods",
    "means",
    "metrics",
    "mlls",
    "models",
    "optim",
    "priors",
    "utils",
    "variational",
    # Classes
    "Module",
    "ExactMarginalLogLikelihood",
    # Functions
    "add_diagonal",
    "add_jitter",
    "dsmm",
    "inv_quad",
    "inv_quad_logdet",
    "pivoted_cholesky",
    "root_decomposition",
    "root_inv_decomposition",
    "solve",
    "sqrt_inv_matmul",
    # Context managers
    "beta_features",
    "settings",
    # Other
    "__version__",
    # Deprecated
    "inv_matmul",
    "logdet",
    "log_normal_cdf",
    "matmul",
]
